#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT

__all__ = ["new"]

from typing import Annotated

import typer

from chango.config import get_chango_instance


def new(
    slug: Annotated[
        str, typer.Option("--slug", "-s", help="The slug of the change note.", show_default=False)
    ],
    edit: Annotated[
        bool,
        typer.Option(
            "--edit/--no-edit",
            "-e/-ne",
            help="Whether to open the change note in the default editor.",
        ),
    ] = True,
) -> None:
    """Create a new change note."""
    change_note = get_chango_instance().build_template_change_note(slug=slug)
    path = get_chango_instance().write_change_note(change_note, version=None)
    typer.echo(f"Created new change note {change_note.file_name}")
    if edit:
        typer.launch(path.as_posix())



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT

__all__ = ["edit"]

from typing import Annotated

import typer

from chango.config import get_chango_instance


def edit(
    uid: Annotated[
        str,
        typer.Argument(
            help="The unique identifier of the change note to edit.", show_default=False
        ),
    ],
) -> None:
    """Edit an existing change note in the default editor."""
    typer.launch(get_chango_instance().scanner.lookup_change_note(uid).file_path.as_posix())



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT

__all__ = ["release"]

import datetime as dtm
from typing import Annotated

import typer

from chango import Version
from chango.config import get_chango_instance

from .utils.types import date as date_callback


def _today() -> dtm.date:
    return dtm.date.today()


def release(
    uid: Annotated[
        str, typer.Option(help="The unique identifier of the version release.", show_default=False)
    ],
    date: Annotated[
        dtm.date,
        typer.Option(
            help="The date of the version release. Defaults to today.",
            parser=date_callback,
            default_factory=_today,
        ),
    ],
) -> None:
    """Release the unreleased changes to a new version."""
    if get_chango_instance().release(Version(uid, date)):
        typer.echo(f"Released version {uid} on {date}")
    else:
        typer.echo("No unreleased changes found.")



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import contextlib
import functools
import sys
from collections.abc import Callable, Generator, Iterator
from contextlib import contextmanager
from pathlib import Path
from typing import ClassVar, override

from pydantic import BaseModel, ConfigDict
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    PyprojectTomlConfigSettingsSource,
)

from chango._utils.types import PathLike


class FrozenModel(BaseModel):
    """A frozen Pydantic model."""

    model_config: ClassVar[ConfigDict] = ConfigDict(
        frozen=True, arbitrary_types_allowed=True, extra="forbid"
    )

    @classmethod
    @contextmanager
    def _unfrozen(cls) -> Generator[None, None, None]:
        original_frozen = cls.model_config["frozen"]
        cls.model_config["frozen"] = False
        try:
            yield
        finally:
            cls.model_config["frozen"] = original_frozen


class TomlSettings(BaseSettings):
    """Example loading values from the table used by default."""

    _SOURCE_FACTORY: Callable[[type[BaseSettings]], PyprojectTomlConfigSettingsSource] = (
        functools.partial(PyprojectTomlConfigSettingsSource)
    )

    @classmethod
    @contextmanager
    def _with_path(cls, path: Path) -> Generator[None, None, None]:
        """Temporarily load from a different path."""
        cls._SOURCE_FACTORY = functools.partial(PyprojectTomlConfigSettingsSource, toml_file=path)
        try:
            yield
        finally:
            cls._SOURCE_FACTORY = functools.partial(PyprojectTomlConfigSettingsSource)

    @classmethod
    @override
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        return init_settings, cls._SOURCE_FACTORY(settings_cls)


@contextlib.contextmanager
def add_sys_path(path: Path | None) -> Iterator[None]:
    """Temporarily add the given path to `sys.path`."""
    if path is None:
        yield
        return

    effective_path = path.resolve()

    try:
        sys.path.insert(0, str(effective_path))
        yield
    finally:
        sys.path.remove(str(effective_path))


def get_pyproject_toml_path(path: PathLike | None) -> Path:
    """Get the path to the pyproject.toml file."""
    effective_path = Path.cwd() if path is None else Path(path).resolve()
    if not effective_path.is_file():
        effective_path = effective_path / "pyproject.toml"
    return effective_path



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
"""This module contains abstract base classes defining the interfaces that the chango package
is build on.
"""

__all__ = ["ChanGo", "ChangeNote", "VersionHistory", "VersionNote", "VersionScanner"]

from ._changenote import ChangeNote
from ._chango import ChanGo
from ._versionhistory import VersionHistory
from ._versionnote import VersionNote
from ._versionscanner import VersionScanner



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
from pathlib import Path

from chango.helpers import change_uid_from_file, ensure_uid


class TestHelpers:
    def test_ensure_uid_none(self):
        assert ensure_uid(None) is None

    def test_ensure_uid_str(self):
        assert ensure_uid("uid") == "uid"

    def test_ensure_uid_obj(self):
        class Obj:
            uid = "uid"

        assert ensure_uid(Obj()) == "uid"

    def test_ensure_uid_obj_prop(self):
        class Obj:
            @property
            def uid(self):
                return "uid"

        assert ensure_uid(Obj()) == "uid"

    def test_change_uid_from_file(self):
        assert change_uid_from_file("slug.uid.md") == "uid"
        assert change_uid_from_file(Path("slug.uid.md")) == "uid"



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import datetime as dtm

import pytest

from chango import Version
from chango.concrete import CommentChangeNote, CommentVersionNote


class TestVersionNote:
    """Since VersionNote is an abstract base class, we are testing with CommentVersionNote as a
    simple implementation.

    Note that we do *not* test abstract methods, as that is the responsibility of the concrete
    implementations.
    """

    @pytest.fixture(
        autouse=True,
        params=[None, Version(uid="1.0.0", date=dtm.date(2024, 1, 1))],
        ids=["None", "Version"],
    )
    def setup(self, request):
        # This is the next best thing to parametrizing __init__ that I could find
        # in reasonable time
        version = request.param
        self.version = version
        self.version_note = CommentVersionNote(version=version)
        self.change_notes = [
            CommentChangeNote(slug=f"slug-{i}", uid=f"uid-{i}", comment=f"comment-{i}")
            for i in range(5)
        ]

    @property
    def change_note(self) -> CommentChangeNote:
        return self.change_notes[0]

    def test_init(self):
        assert self.version_note.version == self.version
        assert len(self.version_note) == 0

    def test_uid(self):
        assert self.version_note.uid == (self.version.uid if self.version else None)

    def test_date(self):
        assert self.version_note.date == (self.version.date if self.version else None)

    @pytest.mark.parametrize("key_type", ["uid", "filename"])
    def test_set_get_del_item(self, key_type):
        key = self.change_note.uid if key_type == "uid" else self.change_note.file_name

        self.version_note[key] = self.change_note
        assert self.version_note[key] == self.change_note
        del self.version_note[key]

        with pytest.raises(KeyError):
            self.version_note[key]

        with pytest.raises(KeyError):
            del self.version_note[self.change_note.uid]

    def test_setitem_warning(self):
        with pytest.warns(
            UserWarning, match="Key 'non-matching-key' does not match change note UID"
        ):
            self.version_note["non-matching-key"] = self.change_note

    @pytest.mark.parametrize("key_type", ["uid", "filename"])
    def test_add_remove_note(self, key_type):
        key = self.change_note.uid if key_type == "uid" else self.change_note.file_name

        self.version_note.add_change_note(self.change_note)
        assert self.version_note[key] == self.change_note
        self.version_note.remove_change_note(self.change_note)

        with pytest.raises(KeyError):
            self.version_note.remove_change_note(self.change_note)

    def test_iter(self):
        for change_note in self.change_notes:
            self.version_note.add_change_note(change_note)

        for i, key in enumerate(self.version_note):
            assert self.version_note[key] is self.change_notes[i]

    def test_len(self):
        assert len(self.version_note) == 0
        for change_note in self.change_notes:
            self.version_note.add_change_note(change_note)
        assert len(self.version_note) == len(self.change_notes)



import inspect
from unittest.mock import MagicMock

import pytest

from chango.concrete import BackwardCompatibleChanGo, BackwardCompatibleVersionScanner
from chango.error import ChanGoError


class TestBackwardCompatibleChanGo:
    @staticmethod
    def build_mocks(
        *,
        chango: tuple[str, list[object]] | None = None,
        scanner: tuple[str, list[object]] | None = None,
    ) -> tuple[MagicMock, list[MagicMock]]:
        if not chango and not scanner:
            raise ValueError("At least one of 'chango' or 'scanner' must be provided.")
        if chango and scanner and len(chango[1]) != len(scanner[1]):
            raise ValueError(
                "The number of chango instances must match the number of scanner instances."
            )
        if (effective_length := len(chango[1] if chango else scanner[1])) < 2:  # noqa PLR2004
            raise ValueError("At least two instances must be provided.")

        def setup_mock(mock, method_name, result_):
            method = getattr(mock, method_name)
            if isinstance(result_, Exception) or (
                inspect.isclass(result_) and issubclass(result_, Exception)
            ):
                method.side_effect = result_
            else:
                method.return_value = result_

        instances = [MagicMock() for _ in range(effective_length)]
        if chango:
            for instance, result in zip(instances, chango[1], strict=False):
                setup_mock(instance, chango[0], result)
        if scanner:
            for instance, result in zip(instances, scanner[1], strict=False):
                setup_mock(instance.scanner, scanner[0], result)

        return instances[0], instances[1:]

    def test_scanner(self):
        chango = BackwardCompatibleChanGo(MagicMock(), [MagicMock(), MagicMock()])
        assert isinstance(chango.scanner, BackwardCompatibleVersionScanner)

    def test_build_template_change_note(self):
        expected_template = object()
        main_instance, legacy_instances = self.build_mocks(
            chango=("build_template_change_note", [expected_template, RuntimeError, RuntimeError]),
            scanner=None,
        )
        chango = BackwardCompatibleChanGo(main_instance, legacy_instances)

        assert chango.build_template_change_note("slug") is expected_template
        main_instance.build_template_change_note.assert_called_once_with("slug", None)
        for legacy_instance in legacy_instances:
            assert not legacy_instance.build_template_change_note.called

        assert chango.build_template_change_note("slug", "uid") is expected_template
        main_instance.build_template_change_note.assert_called_with("slug", "uid")
        for legacy_instance in legacy_instances:
            assert not legacy_instance.build_template_change_note.called

    @pytest.mark.parametrize(
        ("is_available", "version_note", "expected"),
        [
            (
                [True, False, False],
                ["main-version-note", ChanGoError, ChanGoError],
                "main-version-note",
            ),
            (
                [False, True, False],
                [ChanGoError, "legacy-version-note-1", ChanGoError],
                "legacy-version-note-1",
            ),
            (
                [False, False, True],
                [ChanGoError, ChanGoError, "legacy-version-note-2"],
                "legacy-version-note-2",
            ),
        ],
    )
    @pytest.mark.parametrize("version", ["version", None])
    def test_build_version_note(self, is_available, version_note, version, expected):
        main_instance, legacy_instances = self.build_mocks(
            chango=("build_version_note", version_note), scanner=("is_available", is_available)
        )
        chango = BackwardCompatibleChanGo(main_instance, legacy_instances)

        assert chango.build_version_note(version) == expected
        has_returned = False
        for instance, was_available in zip(
            [main_instance, *legacy_instances], is_available, strict=False
        ):
            if not has_returned:
                instance.scanner.is_available.assert_called_once_with(version)
                if was_available:
                    instance.build_version_note.assert_called_once_with(version)
                else:
                    assert not instance.build_version_note.called
                has_returned = has_returned or was_available
            else:
                assert not instance.scanner.is_available.called
                assert not instance.build_version_note.called

    def test_build_version_note_not_found(self):
        main_instance, legacy_instances = self.build_mocks(
            chango=("build_version_note", [ChanGoError, ChanGoError, ChanGoError]),
            scanner=("is_available", [False, False, False]),
        )
        chango = BackwardCompatibleChanGo(main_instance, legacy_instances)

        with pytest.raises(ChanGoError):
            chango.build_version_note("version")

        for instance in [main_instance, *legacy_instances]:
            instance.scanner.is_available.assert_called_once_with("version")
            instance.build_version_note.assert_not_called()

    def test_build_version_history(self):
        expected_history = object()
        main_instance, legacy_instances = self.build_mocks(
            chango=("build_version_history", [expected_history, RuntimeError, RuntimeError]),
            scanner=None,
        )
        chango = BackwardCompatibleChanGo(main_instance, legacy_instances)

        assert chango.build_version_history() is expected_history
        main_instance.build_version_history.assert_called_once_with()
        for legacy_instance in legacy_instances:
            assert not legacy_instance.build_version_history.called

    @pytest.mark.parametrize(
        ("results", "expected"),
        [
            (["main-change-note", ChanGoError, ChanGoError], "main-change-note"),
            ([ChanGoError, "legacy-change-note-1", ChanGoError], "legacy-change-note-1"),
            ([ChanGoError, ChanGoError, "legacy-change-note-2"], "legacy-change-note-2"),
        ],
    )
    def test_load_change_note(self, results, expected):
        main_instance, legacy_instances = self.build_mocks(
            chango=("load_change_note", [expected, ChanGoError, ChanGoError]), scanner=None
        )
        chango = BackwardCompatibleChanGo(main_instance, legacy_instances)

        assert chango.load_change_note("uid") == expected

        has_returned = False
        for instance, result in zip([main_instance, *legacy_instances], results, strict=False):
            if not has_returned:
                instance.load_change_note.assert_called_once_with("uid")
                has_returned = has_returned or not isinstance(result, ChanGoError)
            else:
                assert not instance.load_change_note.called

    def test_load_change_note_not_found(self):
        main_instance, legacy_instances = self.build_mocks(
            chango=("load_change_note", [ChanGoError, ChanGoError, ChanGoError]), scanner=None
        )
        chango = BackwardCompatibleChanGo(main_instance, legacy_instances)

        with pytest.raises(ChanGoError):
            chango.load_change_note("uid")

        for instance in [main_instance, *legacy_instances]:
            instance.load_change_note.assert_called_once_with("uid")

    def test_get_write_directory(self):
        expected_directory = object()
        main_instance, legacy_instances = self.build_mocks(
            chango=("get_write_directory", [expected_directory, RuntimeError, RuntimeError]),
            scanner=None,
        )
        chango = BackwardCompatibleChanGo(main_instance, legacy_instances)

        assert chango.get_write_directory("change_note", "version") is expected_directory
        main_instance.get_write_directory.assert_called_once_with("change_note", "version")
        for legacy_instance in legacy_instances:
            assert not legacy_instance.get_write_directory.called

    def test_build_github_event_change_note(self):
        expected = object()
        main_instance, legacy_instances = self.build_mocks(
            chango=("build_github_event_change_note", [expected, RuntimeError, RuntimeError]),
            scanner=None,
        )
        chango = BackwardCompatibleChanGo(main_instance, legacy_instances)

        call_args = [[("event", "data"), ("event", "data")], [("event",), ("event", None)]]

        for args, expected_args in call_args:
            assert chango.build_github_event_change_note(*args) is expected
            main_instance.build_github_event_change_note.assert_called_with(*expected_args)
            for legacy_instance in legacy_instances:
                assert not legacy_instance.build_github_event_change_note.called



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import subprocess
from pathlib import Path

UTF8 = "utf-8"


class _GitHelper:
    # Alternatives to using subprocess would be using pygit2 or dulwich.
    # However, that would add rather heavy dependencies for a very small part of this library.
    # Let's keep it in mind for a possible future improvement.

    def __init__(self) -> None:
        self.git_available: bool | None = None

    @staticmethod
    def _git_move(src: Path, dst: Path) -> None:
        subprocess.check_call(["git", "mv", str(src), str(dst)])

    @staticmethod
    def _git_add(path: Path) -> None:
        """Add a file to the git index."""
        subprocess.check_call(["git", "add", str(path)])

    @staticmethod
    def _pathlib_move(src: Path, dst: Path) -> None:
        src.rename(dst)

    def move(self, src: Path, dst: Path) -> None:
        if self.git_available is None:
            # We use try-except instead of first checking if git is available
            # because that way we can avoid calling git twice.
            try:
                self._git_move(src, dst)
                self.git_available = True
            except subprocess.CalledProcessError:
                self.git_available = False
                self._pathlib_move(src, dst)

        elif self.git_available:
            self._git_move(src, dst)
        else:
            self._pathlib_move(src, dst)

    def add(self, path: Path) -> None:
        if self.git_available is False:
            return
        if self.git_available:
            self._git_add(path)
        else:
            try:
                self._git_add(path)
                self.git_available = True
            except subprocess.CalledProcessError:
                self.git_available = False


_GIT_HELPER = _GitHelper()


def move_file(src: Path, dst: Path) -> None:
    _GIT_HELPER.move(src, dst)


def try_git_add(path: Path) -> None:
    """Add a file to the git index if git is available."""
    _GIT_HELPER.add(path)



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import importlib
from pathlib import Path
from typing import Annotated, Any, ClassVar, Self, cast

from pydantic import Field
from pydantic_settings import SettingsConfigDict

from chango.abc import ChanGo

from .._utils.config import FrozenModel, TomlSettings, add_sys_path, get_pyproject_toml_path
from .._utils.types import PathLike

__all__ = ["ChanGoConfig", "ChanGoInstanceConfig"]


class ChanGoInstanceConfig(FrozenModel):
    """Data structure for specifying how the :class:`~chango.abc.ChanGo` should be imported for the
    CLI.

    Args:
        name (:obj:`str`): The name of the object to import.
        module (:obj:`str`): The module to import the object from as passed to
            :func:`importlib.import_module`.
        package (:obj:`str` | :obj:`None`, optional): The module to import the object from as
            passed to :func:`importlib.import_module`.

    Attributes:
        name (:obj:`str`): The name of the object to import.
        module (:obj:`str`): The module to import the object from as passed to
            :func:`importlib.import_module`.
        package (:obj:`str` | :obj:`None`): The module to import the object from as passed to
            :func:`importlib.import_module`.
    """

    name: str
    module: Annotated[str, Field(examples=["my_config_module"])]
    package: str | None = None


class ChanGoConfig(FrozenModel, TomlSettings):
    """Data structure for the ChanGos CLI configuration in the ``pyproject.toml`` file.

    Tip:
        Rather than manually creating an instance of this class, use :meth:`load` to load the
        configuration from the ``pyproject.toml`` file.

    Important:
        The attributes of :attr:`chango_instance` will be passed to
        :func:`importlib.import_module` to import the user defined :class:`~chango.abc.ChanGo`
        instance. For this to work,
        the module must be findable by Python, which may depend on your current working directory
        and the Python path. It can help to set :paramref:`sys_path` accordingly.
        Please evaluate the security implications of this before setting it.

    Keyword Args:
        sys_path (:class:`~pathlib.Path`, optional): A path to *temporarily* add to the system
            path before importing the module.

            Example:
                To add the current working directory to the system path, set this to ``.``.

            Caution:
                Since this class is usually loaded via :meth:`load`, the path is resolved relative
                to the ``pyproject.toml`` file path. If the path is absolute, it will be used as
                is.
                When instantiating this class manually, the path is resolved relative to the
                current working directory.

        chango_instance (:class:`~chango.config.ChanGoInstanceConfig`): Specification of how the
            :class:`~chango.abc.ChanGo` instance to use in the CLI is imported.

    Attributes:
        sys_path (:class:`~pathlib.Path` | None): The path to *temporarily* add to the system path
            before importing the module. If the path is not absolute, it will considered as
            relative to the current working directory.
        chango_instance (:class:`~chango.config.ChanGoInstanceConfig`): The instance of
            :class:`~chango.abc.ChanGo` to use in the CLI.
    """

    model_config: ClassVar[SettingsConfigDict] = SettingsConfigDict(
        pyproject_toml_table_header=("tool", "chango"), extra="ignore"
    )

    sys_path: Path | None = None
    chango_instance: Annotated[
        ChanGoInstanceConfig,
        Field(examples=[ChanGoInstanceConfig(name="chango_instance", module="my_config_module")]),
    ]

    @classmethod
    def load(cls, path: PathLike | None = None) -> Self:
        """Load the :class:`~chango.config.ChanGoConfig` from the ``pyproject.toml`` file.

        Tip:
           If the specification of :attr:`sys_path` is relative, it will be resolved relative to
           the :paramref:`path` parameter by this method.

        Keyword Args:
            path (:class:`~pathlib.Path` | None): The path to the ``pyproject.toml`` file. The
                path resolution works as follows:

                * If ``path`` is ``None``, the current working directory is used.
                * If ``path`` is absolute, it is used as is. Relative paths are resolved relative
                  to the current working directory.
                * If the path does not point to a file, it is assumed to be a directory and the
                  file name ``pyproject.toml`` is appended.

        Returns:
            :class:`~chango.config.ChanGoConfig`: The loaded configuration.
        """
        pyproject_toml_path = get_pyproject_toml_path(path)
        with cls._with_path(get_pyproject_toml_path(path)):
            obj = cls()  # type: ignore[call-arg]
            if obj.sys_path is None or obj.sys_path.is_absolute():
                return obj
            with obj._unfrozen():
                obj.sys_path = (pyproject_toml_path.parent / obj.sys_path).resolve()
            return obj

    def import_chango_instance(self) -> ChanGo[Any, Any, Any, Any]:
        """Import the :class:`~chango.abc.ChanGo` instance specified in :attr:`chango_instance`.
        This considers the :attr:`sys_path` attribute to temporarily add a path to the system path.

        Returns:
            :class:`~chango.abc.ChanGo`: The imported :class:`~chango.abc.ChanGo` instance.
        """
        with add_sys_path(self.sys_path):
            return cast(
                "ChanGo",
                getattr(
                    importlib.import_module(
                        self.chango_instance.module, self.chango_instance.package
                    ),
                    self.chango_instance.name,
                ),
            )



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import ClassVar, Self

import shortuuid

from ..error import ValidationError

_short_uuid = shortuuid.ShortUUID()


def random_uid() -> str:
    return _short_uuid.uuid()


@dataclass(frozen=True)
class FileName:
    slug: str
    uid: str = field(default_factory=random_uid)

    SEPARATOR: ClassVar[str] = "."

    def __post_init__(self) -> None:
        if self.SEPARATOR in self.slug:
            raise ValidationError(f"slug must not contain {self.SEPARATOR!r}")

    @classmethod
    def from_string(cls, string: str) -> Self:
        try:
            slug, uid, _ = string.split(cls.SEPARATOR)
        except ValueError as exc:
            raise ValidationError(f"invalid filename: {string!r}") from exc

        return cls(slug, uid)

    def to_string(self, extension: str) -> str:
        return self.SEPARATOR.join([self.slug, self.uid, extension])



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
#
#  SPDX-License-Identifier: MIT
#
#  SPDX-License-Identifier: MIT
import contextlib
from collections.abc import Collection
from typing import override

from .._changenoteinfo import ChangeNoteInfo
from .._utils.types import VUIDInput
from .._version import Version
from ..abc import VersionScanner
from ..error import ChanGoError


class BackwardCompatibleVersionScanner(VersionScanner):
    """An Implementation of the :class:`~chango.abc.VersionScanner` interface that wraps multiple
    other implementations of :class:`~chango.abc.VersionScanner`.
    The purpose of this class is to ease transition between different version note formats in
    a project.

    Warning:
        This assumes that the versions available for each of the scanners are mutually exclusive,
        i.e. no two scanners can return the same version.

    Tip:
        Use together with :class:`~chango.concrete.BackwardCompatibleChanGo`.

    Args:
        scanners (Collection[:class:`~chango.abc.VersionScanner`]): The scanners to wrap.
    """

    def __init__(self, scanners: Collection[VersionScanner]):
        self._scanners = tuple(scanners)

    @override
    def is_available(self, uid: VUIDInput) -> bool:
        return any(scanner.is_available(uid) for scanner in self._scanners)

    @override
    def has_unreleased_changes(self) -> bool:
        return any(scanner.has_unreleased_changes() for scanner in self._scanners)

    @override
    def get_latest_version(self) -> Version:
        """Implementation of :meth:`chango.abc.VersionScanner.get_latest_version`.

        Important:
            The newest version is determined by the date of the version, not the order in which
            the scanners were passed to the constructor.

        Returns:
            :class:`~chango.Version`: The latest version
        """
        versions = []
        for scanner in self._scanners:
            with contextlib.suppress(ChanGoError):
                versions.append(scanner.get_latest_version())
        if not versions:
            raise ChanGoError("No versions available.")
        return max(versions, key=lambda v: v.date)

    @override
    def get_available_versions(
        self, start_from: VUIDInput = None, end_at: VUIDInput = None
    ) -> tuple[Version, ...]:
        return tuple(
            version
            for scanner in self._scanners
            for version in scanner.get_available_versions(start_from, end_at)
        )

    @override
    def lookup_change_note(self, uid: str) -> ChangeNoteInfo:
        """Lookup a change note with the given identifier.

        Args:
            uid (:obj:`str`): The unique identifier or file name of the change note to lookup

        Returns:
            :class:`chango.ChangeNoteInfo`: The metadata about the change note specifying the file
                path and version it belongs to.
        """
        for scanner in self._scanners:
            with contextlib.suppress(ChanGoError):
                return scanner.lookup_change_note(uid)
        raise ChanGoError(f"Change note '{uid}' not available.")

    @override
    def get_changes(self, uid: VUIDInput) -> tuple[str, ...]:
        for scanner in self._scanners:
            with contextlib.suppress(ChanGoError):
                return scanner.get_changes(uid)
        raise ChanGoError(f"Version '{uid}' not available.")

    @override
    def invalidate_caches(self) -> None:
        for scanner in self._scanners:
            scanner.invalidate_caches()



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
from pathlib import Path

import pytest

from chango import ChangeNoteInfo, Version


class TestChangeNoteInfo:
    def test_init(self):
        v = Version("1.2.3", None)
        cni = ChangeNoteInfo(uid="123", version=v, file_path=Path("/path/to/file"))
        assert cni.uid == "123"
        assert cni.version == v
        assert cni.file_path == Path("/path/to/file")

    def test_frozen(self):
        v = Version("1.2.3", None)
        cni = ChangeNoteInfo(uid="123", version=v, file_path=Path("/path/to/file"))
        with pytest.raises(AttributeError):
            cni.uid = "124"
        with pytest.raises(AttributeError):
            cni.version = Version("1.2.4", None)
        with pytest.raises(AttributeError):
            cni.file_path = Path("/new/path/to/file")



# SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
# SPDX-License-Identifier: MIT

__all__ = [
    "ChangeNoteInfo",
    "Version",
    "__version__",
    "abc",
    "action",
    "concrete",
    "config",
    "constants",
    "error",
    "helpers",
]

from . import __about__, abc, action, concrete, config, constants, error, helpers
from ._changenoteinfo import ChangeNoteInfo
from ._version import Version

#: :obj:`str`: The version of the ``chango`` library as string
__version__: str = __about__.__version__



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
from chango.config import ChanGoInstanceConfig


class TestChanGoInstanceConfig:
    def test_init(self):
        config = ChanGoInstanceConfig(name="name", module="module", package="package")
        assert config.name == "name"
        assert config.module == "module"
        assert config.package == "package"

    def test_init_required(self):
        config = ChanGoInstanceConfig(name="name", module="module")
        assert config.name == "name"
        assert config.module == "module"
        assert config.package is None



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT


# INFO:
# Best reference for how use sphinx testing so far is
# https://github.com/sphinx-doc/sphinx/issues/7008

pytest_plugins = ["sphinx.testing.fixtures"]



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import traceback
import unittest.mock
from collections.abc import Mapping, Sequence
from typing import IO, Any, Literal

import pytest
from click.testing import Result
from typer import Typer
from typer.testing import CliRunner

import chango
from chango._cli import app as chango_app


class CLIResult:
    def __init__(self, result: Result) -> None:
        self._result = result

    def __getattr__(self, name: str) -> Any:
        # only called if attribute is not found in self
        return getattr(self._result, name)

    def check_exit_code(self, expected_code: int = 0) -> Literal[True]:
        text = f"stdout:\n{self.stdout}"
        if self.exception:
            text += f"\nexception:\n{self.exception}"
            text += f"\ntraceback:\n{'\n'.join(traceback.format_exception(*self.exc_info))}"

        if self.exit_code != expected_code:
            raise AssertionError(
                f"Expected exit code {expected_code}, got {self.exit_code}\n{text}"
            )
        return True


class ReuseCliRunner(CliRunner):
    def __init__(self, app: Typer, *args: Any, **kwargs: Any) -> None:
        self.app = app
        # For easier testing, disable rich markup mode
        self.app.rich_markup_mode = None
        super().__init__(*args, **kwargs)

    def invoke(
        self,
        args: str | Sequence[str] | None = None,
        input: bytes | str | IO[Any] | None = None,
        env: Mapping[str, str] | None = None,
        catch_exceptions: bool = True,
        color: bool = False,
        **extra: Any,
    ) -> CLIResult:
        return CLIResult(
            super().invoke(
                self.app,
                args=args,
                input=input,
                env=env,
                catch_exceptions=catch_exceptions,
                color=color,
                **extra,
            )
        )


@pytest.fixture(scope="session")
def cli():
    return ReuseCliRunner(chango_app)


@pytest.fixture
def mock_chango_instance(monkeypatch):
    chango_config = unittest.mock.MagicMock()
    monkeypatch.setattr(chango.config.ChanGoConfig, "load", lambda *_, **__: chango_config)

    yield chango_config.import_chango_instance()

    # This is required to ensure that each test gets a new instance of the mock
    chango.config.get_chango_instance.cache_clear()



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
from pathlib import Path
from typing import Protocol, overload

from ._utils.filename import FileName
from ._utils.types import PathLike

__all__ = ["change_uid_from_file", "ensure_uid"]


def change_uid_from_file(file: PathLike) -> str:
    """Get the change note identifier from a file name or path.

    Args:
        file (:obj:`str` | :class:`pathlib.Path`): The file name or path to get the identifier
            from.

    Returns:
        :obj:`str`: The :attr:`~chango.abc.ChangeNote.uid` of the change note.
    """
    if isinstance(file, Path):
        return change_uid_from_file(file.name)
    return FileName.from_string(file).uid


class _UIDProtocol(Protocol):
    uid: str


class _UIDPropProtocol(Protocol):
    @property
    def uid(self) -> str: ...


@overload
def ensure_uid(obj: _UIDProtocol | _UIDPropProtocol) -> str: ...


@overload
def ensure_uid(obj: None) -> None: ...


@overload
def ensure_uid(obj: str) -> str: ...


def ensure_uid(obj: _UIDProtocol | _UIDPropProtocol | str | None) -> str | None:
    """Extract the unique identifier of an object. Input of type :obj:`str` and :obj:`None`
    is returned unchanged.

    Args:
        obj (:obj:`str` | :obj:`None` | has ``uid``): An object that either has a string attribute
            ``uid`` (e.g. :class:`~chango.abc.ChangeNote` or :class:`~chango.Version`), is a
            :obj:`str` or is :obj:`None`.

    Returns:
        :obj:`str` | :obj:`None`: The extracted UID if available and :obj:`None` else.
    """
    if obj is None:
        return None
    if isinstance(obj, str):
        return obj
    return obj.uid



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import datetime as dtm

import pytest

from chango import Version
from chango.concrete import CommentChangeNote, CommentVersionNote, HeaderVersionHistory


class TestVersionHistory:
    """Since VersionHistory is an abstract base class, we are testing with CommentVersionHistory as
    a simple implementation.

    Note that we do *not* test abstract methods, as that is the responsibility of the concrete
    implementations.
    """

    @pytest.fixture(autouse=True)
    def setup(self):
        self.version_notes = []
        for i in range(5):
            version_note = CommentVersionNote(
                version=Version(f"1.0.{i}", date=dtm.date(2024, 1, i + 1))
            )
            for j in range(5):
                version_note.add_change_note(
                    CommentChangeNote(
                        slug=f"1-0-{i}-slug-{j}",
                        uid=f"1-0-{i}-uid-{j}",
                        comment=f"1-0-{i}-comment-{j}",
                    )
                )
            self.version_notes.append(version_note)

        self.version_note = self.version_notes[0]
        self.version_history = HeaderVersionHistory()

    @pytest.fixture(params=["uid", "Version", "None"])
    def key(self, request):
        if request.param == "uid":
            return self.version_note.uid
        if request.param == "Version":
            return self.version_note.version
        return None

    def test_set_get_del_item(self, key):
        version_note = self.version_note if key else CommentVersionNote(None)
        self.version_history[key] = version_note
        assert self.version_history[key] == version_note
        del self.version_history[key]

        with pytest.raises(KeyError):
            self.version_history[key]

        with pytest.raises(KeyError):
            del self.version_history[self.version_note.uid]

    def test_setitem_warning(self):
        with pytest.warns(
            UserWarning, match="Key 'non-matching-key' does not match version note UID"
        ):
            self.version_history["non-matching-key"] = self.version_note

    def test_add_remove_note(self, key):
        if key is None:
            pytest.skip("Not relevant for None key")

        self.version_history.add_version_note(self.version_note)
        assert self.version_history[key] == self.version_note
        self.version_history.remove_version_note(self.version_note)

        with pytest.raises(KeyError):
            self.version_history.remove_version_note(self.version_note)

    def test_iter(self):
        for version_note in self.version_notes:
            self.version_history.add_version_note(version_note)

        for i, key in enumerate(self.version_history):
            assert self.version_history[key] is self.version_notes[i]

    def test_len(self):
        assert len(self.version_history) == 0
        for version_note in self.version_notes:
            self.version_history.add_version_note(version_note)
        assert len(self.version_history) == len(self.version_notes)



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT

import datetime as dtm

import pytest
import shortuuid

from chango import Version
from chango.action import ChanGoActionData, ParentPullRequest
from chango.concrete import (
    CommentChangeNote,
    CommentVersionNote,
    DirectoryChanGo,
    DirectoryVersionScanner,
    HeaderVersionHistory,
)
from chango.concrete.sections import (
    GitHubSectionChangeNote,
    PullRequest,
    Section,
    SectionVersionNote,
)
from chango.error import ChanGoError
from chango.helpers import ensure_uid
from tests.auxil.files import data_path


@pytest.fixture(scope="module")
def scanner() -> DirectoryVersionScanner:
    return DirectoryVersionScanner(TestDirectoryChango.DATA_ROOT, "unreleased")


@pytest.fixture(scope="module")
def chango(scanner) -> DirectoryChanGo:
    return DirectoryChanGo(
        change_note_type=CommentChangeNote,
        version_note_type=CommentVersionNote,
        version_history_type=HeaderVersionHistory,
        scanner=scanner,
    )


DummySectionChangeNote = GitHubSectionChangeNote.with_sections(
    [
        Section(uid="req_0", title="req_0", is_required=True),
        Section(uid="opt_0", title="opt_0", is_required=False),
        Section(uid="opt_1", title="opt_1", is_required=False),
    ]
)


@pytest.fixture(scope="module")
def section_scanner() -> DirectoryVersionScanner:
    return DirectoryVersionScanner(TestDirectoryChango.SECTION_DATA_ROOT, "unreleased")


@pytest.fixture(scope="module")
def section_chango(section_scanner) -> DirectoryChanGo:
    return DirectoryChanGo(
        change_note_type=DummySectionChangeNote,
        version_note_type=SectionVersionNote,
        version_history_type=HeaderVersionHistory,
        scanner=section_scanner,
    )


class TestDirectoryChango:
    DATA_ROOT = data_path("directoryversionscanner")
    SECTION_DATA_ROOT = data_path("directoryversionscanner-sections")

    def test_init_basic(self, chango, scanner):
        assert chango.directory_format == "{uid}_{date}"
        assert chango.change_note_type == CommentChangeNote
        assert chango.version_note_type == CommentVersionNote
        assert chango.version_history_type == HeaderVersionHistory
        assert chango.scanner is scanner

    def test_init_custom_format(self, scanner):
        chango = DirectoryChanGo(
            change_note_type=CommentChangeNote,
            version_note_type=CommentVersionNote,
            version_history_type=HeaderVersionHistory,
            scanner=scanner,
            directory_format="{uid} custom {date}",
        )
        assert chango.directory_format == "{uid} custom {date}"

    @pytest.mark.parametrize("uid", [None, "uid"])
    def test_build_template_change_note(self, chango, uid):
        note = chango.build_template_change_note("slug", uid)
        assert isinstance(note, CommentChangeNote)
        assert note.slug == "slug"
        if uid is not None:
            assert note.uid == uid
        else:
            assert isinstance(note.uid, str)
            assert len(note.uid) == len(shortuuid.ShortUUID().uuid())

    class TestBuildGitHubEeventChangeNote:
        def test_basic(self, chango):
            event = {
                "pull_request": {
                    "html_url": "https://example.com/pull/42",
                    "number": 42,
                    "title": "example title",
                }
            }
            note = chango.build_github_event_change_note(event)
            assert isinstance(note, CommentChangeNote)
            assert isinstance(note.uid, str)
            assert len(note.uid) == len(shortuuid.ShortUUID().uuid())

        @pytest.mark.parametrize(
            "data",
            [
                pytest.param(None, id="None"),
                pytest.param({"some": "dict"}, id="dict"),
                pytest.param(
                    ChanGoActionData(parent_pull_request=None, linked_issues=None),
                    id="ChanGoActionData without parent PR",
                ),
            ],
        )
        def test_section_change_note_early_exit(self, chango, data):
            event = {
                "pull_request": {
                    "html_url": "https://example.com/pull/42",
                    "number": 42,
                    "title": "example title",
                }
            }
            note = chango.build_github_event_change_note(event, data)
            assert isinstance(note, CommentChangeNote)
            assert isinstance(note.uid, str)
            assert len(note.uid) == len(shortuuid.ShortUUID().uuid())

        def test_section_change_note_no_existing(self, section_chango):
            event = {
                "pull_request": {
                    "html_url": "https://example.com/pull/42",
                    "number": 33,
                    "title": "example title",
                    "user": {"login": "author"},
                }
            }
            data = ChanGoActionData(
                parent_pull_request=ParentPullRequest(
                    url="https://example.com/pull/43",
                    number=50,
                    title="example title",
                    state="OPEN",
                    author_login="author",
                ),
                linked_issues=None,
            )
            note = section_chango.build_github_event_change_note(event, data)
            assert isinstance(note, DummySectionChangeNote)
            assert note.slug == "0033"
            assert note.pull_requests == (
                PullRequest(uid="33", author_uids=("author",), closes_threads=()),
            )
            assert note.req_0 == "example title"

        @pytest.mark.parametrize("modifies", [True, False])
        def test_section_change_note_with_existing(self, section_chango, modifies):
            event = {
                "pull_request": {
                    "html_url": "https://example.com/pull/42",
                    "number": 101,
                    "title": "example title" if modifies else "req_0",
                    "user": {"login": "author"},
                }
            }
            note = section_chango.build_github_event_change_note(event, None)
            if not modifies:
                assert note is None
                return

            assert isinstance(note, DummySectionChangeNote)
            assert note.slug == "0101"
            assert note.uid == "abcuc4HVWKycYccXAcM3xc"
            assert note.pull_requests == (
                PullRequest(uid="101", author_uids=("author",), closes_threads=()),
            )
            assert note.req_0 == "example title"

        def test_section_change_note_with_existing_parent(self, section_chango, monkeypatch):
            def get_sections(*args, **kwargs):  # noqa: ARG001
                return {"req_0", "opt_0"}

            monkeypatch.setattr(DummySectionChangeNote, "get_sections", get_sections)

            event = {
                "pull_request": {
                    "html_url": "https://example.com/pull/42",
                    "number": 45,
                    "title": "example title",
                    "user": {"login": "author"},
                }
            }
            data = ChanGoActionData(
                parent_pull_request=ParentPullRequest(
                    url="https://example.com/pull/43",
                    number=43,
                    title="example title",
                    state="OPEN",
                    author_login="author",
                ),
                linked_issues=None,
            )
            note = section_chango.build_github_event_change_note(event, data)

            assert isinstance(note, DummySectionChangeNote)
            assert note.slug == "0043"
            assert note.pull_requests == (
                PullRequest(
                    uid="43",
                    author_uids=("parent_author",),
                    closes_threads=("existing_thread1", "existing_thread2"),
                ),
                PullRequest(uid="45", author_uids=("author",), closes_threads=()),
            )
            assert note.req_0 == "existing_req_0\nexample title"
            assert note.opt_0 == "example title"
            assert note.opt_1 == "existing_opt_1"

    def test_build_version_note_version(self, chango):
        version = Version("uid", dtm.date.today())
        note = chango.build_version_note(version)
        assert isinstance(note, CommentVersionNote)
        assert note.version == version

    def test_build_version_note_none(self, chango):
        note = chango.build_version_note(None)
        assert isinstance(note, CommentVersionNote)
        assert note.version is None

    def test_build_version_note_sections(self, section_chango):
        note = section_chango.build_version_note(None)
        assert isinstance(note, SectionVersionNote)
        assert note.version is None

    def test_build_version_history(self, chango):
        history = chango.build_version_history()
        assert isinstance(history, HeaderVersionHistory)

    @pytest.mark.parametrize("idx", [1, 2, 3])
    def test_load_change_note(self, chango, idx):
        uid = f"uid_1-{idx}_0"
        change_note = chango.load_change_note(uid)
        assert isinstance(change_note, CommentChangeNote)
        assert change_note.uid == uid
        assert change_note.slug == "comment-change-note"
        path = (
            self.DATA_ROOT / f"1.{idx}_2024-01-0{idx}" / f"comment-change-note.uid_1-{idx}_0.txt"
        )
        assert change_note.comment == path.read_text()

    @pytest.mark.parametrize(
        "version",
        [
            None,
            "1.1",
            Version("1.1", dtm.date(2024, 1, 1)),
            "1.2",
            Version("1.2", dtm.date(2024, 1, 2)),
        ],
    )
    @pytest.mark.parametrize(
        "change_note", ["str_change_note", CommentChangeNote.build_template("slug", "uid")]
    )
    def test_get_write_directory(self, chango, version, change_note):
        if version is None:
            expected_path = chango.scanner.unreleased_directory
        else:
            version_uid = ensure_uid(version)
            day = int(version_uid.split(".")[-1])
            expected_path = chango.scanner.base_directory / f"{version_uid}_2024-01-0{day}"

        path_existed = expected_path.is_dir()

        try:
            path = chango.get_write_directory(change_note, version)
            assert path.is_dir()
        finally:
            # Clean up
            if (expected_path is not None) and (not path_existed):
                expected_path.rmdir()

    @pytest.mark.parametrize(
        "change_note", ["str_change_note", CommentChangeNote.build_template("slug", "uid")]
    )
    def test_get_write_directory_new_version(self, chango, change_note):
        version = Version("new-version", dtm.date(2024, 1, 17))
        expected_path = chango.scanner.base_directory / "new-version_2024-01-17"
        try:
            path = chango.get_write_directory(change_note, version)
            assert path.is_dir()
            assert path == expected_path
        finally:
            # Clean up
            expected_path.rmdir()

    @pytest.mark.parametrize(
        "change_note", ["str_change_note", CommentChangeNote.build_template("slug", "uid")]
    )
    def test_get_write_directory_new_str_version(self, chango, change_note):
        with pytest.raises(ChanGoError, match=r"'new-version' not available."):
            chango.get_write_directory(change_note, "new-version")



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT

import pytest
from pydantic import ValidationError

from chango.action import ChanGoActionData, LinkedIssue, ParentPullRequest


@pytest.fixture(scope="module")
def parent_pull_request():
    return ParentPullRequest(
        number=1, author_login="author", title="title", url="http://example.com", state="OPEN"
    )


@pytest.fixture(scope="module")
def linked_issue():
    return LinkedIssue(number=1, title="title", labels=["label1", "label2"], issue_type="type")


class TestParentPullRequest:
    def test_init_basic(self, parent_pull_request):
        assert parent_pull_request.number == 1
        assert parent_pull_request.author_login == "author"
        assert parent_pull_request.title == "title"
        assert str(parent_pull_request.url) == "http://example.com/"
        assert parent_pull_request.state == "OPEN"

    def test_frozen(self, parent_pull_request):
        with pytest.raises(ValidationError, match="frozen"):
            parent_pull_request.number = 2

    def test_invalid_url(self):
        with pytest.raises(ValidationError, match=r"input_value='example.com'"):
            ParentPullRequest(
                number=1, author_login="author", title="title", url="example.com", state="open"
            )

    def test_invalid_state(self):
        with pytest.raises(ValidationError, match="input_value='invalid'"):
            ParentPullRequest(
                number=1,
                author_login="author",
                title="title",
                url="http://example.com",
                state="invalid",
            )


class TestLinkedIssue:
    def test_init_basic(self, linked_issue):
        assert linked_issue.number == 1
        assert linked_issue.title == "title"
        assert linked_issue.labels == ("label1", "label2")
        assert linked_issue.issue_type == "type"

    def test_init_no_optional(self):
        linked_issue = LinkedIssue(number=1, title="title", labels=None)
        assert linked_issue.labels is None
        assert linked_issue.issue_type is None

    def test_frozen(self, linked_issue):
        with pytest.raises(ValidationError, match="frozen"):
            linked_issue.number = 2

    def test_init_none_labels(self):
        linked_issue = LinkedIssue(number=1, title="title", labels=None, issue_type=None)
        assert linked_issue.labels is None
        assert linked_issue.issue_type is None


class TestChanGoActionData:
    def test_init_basic(self, parent_pull_request, linked_issue):
        data = ChanGoActionData(
            parent_pull_request=parent_pull_request, linked_issues=[linked_issue]
        )
        assert data.parent_pull_request == parent_pull_request
        assert data.linked_issues == (linked_issue,)

    def test_init_none(self):
        data = ChanGoActionData(parent_pull_request=None, linked_issues=None)
        assert data.parent_pull_request is None
        assert data.linked_issues is None

    def test_frozen(self, parent_pull_request, linked_issue):
        data = ChanGoActionData(
            parent_pull_request=parent_pull_request, linked_issues=[linked_issue]
        )
        with pytest.raises(ValidationError, match="frozen"):
            data.parent_pull_request = None

        with pytest.raises(ValidationError, match="frozen"):
            data.linked_issues = None



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT

__all__ = ["app"]

from pathlib import Path
from typing import Annotated

import tomlkit
import typer
from pydantic import ValidationError
from rich import print as rprint
from rich.markdown import Markdown

from .._utils.config import get_pyproject_toml_path
from ..config import ChanGoConfig

app = typer.Typer(help="Show or verify the configuration of the chango CLI.")

_PATH_ANNOTATION = Annotated[
    Path | None,
    typer.Option(
        help=(
            "The path to the [code]pyproject.toml[/code] file. "
            "Input behavior as for [code]chango.config.ChanGoConfig.load[/code]."
        )
    ),
]


@app.callback(rich_help_panel="Meta Functionality")
def callback(context: typer.Context, path: _PATH_ANNOTATION = None) -> None:
    effective_path = get_pyproject_toml_path(path)

    if not effective_path.exists():
        raise typer.BadParameter(f"File not found: {effective_path}")

    context.obj = {"path": effective_path}

    try:
        toml_data = tomlkit.load(context.obj["path"].open("rb"))
    except Exception as exc:
        raise typer.BadParameter(f"Failed to parse the configuration file: {exc}") from exc

    try:
        context.obj["data"] = toml_data["tool"]["chango"]  # type: ignore[index]
    except KeyError as exc:
        raise typer.BadParameter(
            "No configuration found for chango in the configuration file."
        ) from exc


@app.command()
def show(context: typer.Context) -> None:
    """Show the configuration."""
    string = f"""
Showing the configuration of the chango CLI as configured in ``{context.obj["path"]}``.
```toml
{tomlkit.dumps(context.obj["data"])}
```
    """
    rprint(Markdown(string))


@app.command()
def validate(context: typer.Context) -> None:
    """Validate the configuration."""
    try:
        config = ChanGoConfig.load(context.obj["path"])
    except ValidationError as exc:
        raise typer.BadParameter(
            f"Validation of config file at {context.obj['path']} failed:\n{exc}"
        ) from exc

    try:
        config.import_chango_instance()
    except ImportError as exc:
        raise typer.BadParameter(
            f"Config file at {context.obj['path']} is valid "
            f"but importing the ChanGo instance failed:\n{exc}"
        ) from exc

    rprint(f"The configuration in [code]{context.obj['path']}[/code] is valid.")



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import abc
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional

from .._utils.files import UTF8, move_file, try_git_add
from .._utils.types import VUIDInput
from ..action import ChanGoActionData
from ._changenote import ChangeNote
from ._versionhistory import VersionHistory
from ._versionnote import VersionNote
from ._versionscanner import VersionScanner

if TYPE_CHECKING:
    from .. import Version


class ChanGo[VST: VersionScanner, VHT: VersionHistory, VNT: VersionNote, CNT: ChangeNote](abc.ABC):
    """Abstract base class for loading :class:`~chango.abc.ChangeNote`,
    :class:`~chango.abc.VersionNote` and :class:`~chango.abc.VersionHistory` objects as well
    as writing :class:`~chango.abc.ChangeNote` objects.
    This class holds the main interface for interacting with the version history and change notes.
    """

    @property
    @abc.abstractmethod
    def scanner(self) -> VST:
        """:class:`VST <typing.TypeVar>`: The :class:`~chango.abc.VersionScanner` used by this
        instance.
        """

    @abc.abstractmethod
    def build_template_change_note(self, slug: str, uid: str | None = None) -> CNT:
        """Build a template change note for the concrete change note type.

        Tip:
            This will be used to create a new change note in the CLI.

        Args:
            slug (:obj:`str`): The slug to use for the change note.
            uid (:obj:`str`, optional): The unique identifier for the change note or :obj:`None`
                to generate a random one.

        Returns:
            :class:`CNT <typing.TypeVar>`:The :class:`~chango.abc.ChangeNote` object.
        """

    @abc.abstractmethod
    def build_version_note(self, version: Optional["Version"]) -> VNT:
        """Build a new empty version note.

        Args:
            version (:class:`~chango.Version` | :obj:`None`): The version of the
                software project this note is for. May be :obj:`None` if the version is not yet
                released.

        Returns:
            :class:`VNT <typing.TypeVar>`: The :class:`~chango.abc.VersionNote` object.
        """

    @abc.abstractmethod
    def build_version_history(self) -> VHT:
        """:class:`VHT <typing.TypeVar>`: Build a new empty version history."""

    @abc.abstractmethod
    def load_change_note(self, uid: str) -> CNT:
        """Load a change note with the given identifier.

        Args:
            uid (:obj:`str`): The unique identifier or file name of the change note to load.

        Returns:
            :class:`CNT <typing.TypeVar>`: The :class:`~chango.abc.ChangeNote` object.

        Raises:
            ~chango.error.ChanGoError: If the change note with the given identifier is not
                available.
        """

    @abc.abstractmethod
    def get_write_directory(self, change_note: CNT | str, version: VUIDInput) -> Path:
        """Determine the directory to write a change note to.

        Important:
            * It should be ensured that the directory exists.
            * The :paramref:`version` does *not* need to be already available. In that case, it's
              expected that :paramref:`version` is of type :class:`~chango.Version`.

        Args:
            change_note (:class:`CNT <typing.TypeVar>` | :obj:`str`): The change note to write or
                its UID.
            version (:class:`~chango.Version` | :obj:`str` | :obj:`None`): The version the change
                note belongs to. Maybe be :obj:`None` if the change note is not yet released.

        Returns:
            :class:`pathlib.Path`: The directory to write the change note to.

        Raises:
            ~chango.error.ChanGoError: If the :paramref:`version` is a :obj:`str` but not yet
                available.
        """

    def build_github_event_change_note(
        self, event: dict[str, Any], data: dict[str, Any] | ChanGoActionData | None = None
    ) -> CNT | None:
        """Build a change note from a GitHub event.

        Important:
            This is an optional method and by default raises a :class:`NotImplementedError`.
            Implement this method if you want to automatically create change notes based on
            GitHub events.

        Tip:
            This method is useful for automatically creating change note drafts in GitHub actions
            to ensure that each pull request has documented changes.

            .. seealso:: :ref:`action`

        Args:
            event (Dict[:obj:`str`, :obj:`~typing.Any`]): The GitHub event data. This should be one
              of the `events that trigger workflows <ettw>`_. The event is represented as a
              JSON dictionary.
            data (Dict[:obj:`str`, :obj:`~typing.Any`] | :class:`chango.action.ChanGoActionData`, \
               optional): Additional data that may be required to build the change note.

        Returns:
            :class:`CNT <typing.TypeVar>` | :obj:`None`: The change note or :obj:`None` if no
                change note should be created (e.g., if a change note is already available)
                for the change.

        Raises:
            NotImplementedError: If the method is not implemented.

        .. _ettw: https://docs.github.com/en/actions/writing-workflows/\
            choosing-when-your-workflow-runs/events-that-trigger-workflows
        """
        raise NotImplementedError

    def write_change_note(
        self, change_note: CNT, version: VUIDInput, encoding: str = UTF8
    ) -> Path:
        """Write a change note to disk.

        Important:
            The :paramref:`version` does *not* need to be already available. In that case, it's
            expected that :paramref:`version` is of type :class:`~chango.Version`.

        Tip:
            This method calls :meth:`chango.abc.VersionScanner.invalidate_caches` after writing
            the change note to disk.

        Args:
            change_note (:class:`CNT <typing.TypeVar>` | :obj:`str`): The change note to write.
            version (:class:`~chango.Version` | :obj:`str` | :obj:`None`): The version the change
                note belongs to. Maybe be :obj:`None` if the change note is not yet released.
            encoding (:obj:`str`): The encoding to use for writing.

        Returns:
            :class:`pathlib.Path`: The file path the change note was written to.

        Raises:
            ~chango.error.ChanGoError: If the :paramref:`version` is a :obj:`str` but not yet
                available.
        """
        path = change_note.to_file(
            directory=self.get_write_directory(change_note=change_note, version=version),
            encoding=encoding,
        )
        try_git_add(path)
        self.scanner.invalidate_caches()
        return path

    def load_version_note(self, version: VUIDInput) -> VNT:
        """Load a version note.

        Args:
            version (:class:`~chango.Version` | :obj:`str` | :obj:`None`): The version of the
                version note to load or the corresponding uid. May be :obj:`None` if the version is
                not yet released.

        Returns:
            :class:`VNT <typing.TypeVar>`: The loaded :class:`~chango.abc.VersionNote`.

        Raises:
            ~chango.error.ChanGoError: If the version is not available.
        """
        changes = self.scanner.get_changes(version)
        version_obj = self.scanner.get_version(version) if isinstance(version, str) else version
        version_note = self.build_version_note(version=version_obj)
        for change in changes:
            version_note.add_change_note(self.load_change_note(change))

        return version_note

    def load_version_history(self, start_from: VUIDInput = None, end_at: VUIDInput = None) -> VHT:
        """Load the version history.

        Important:
            By default, unreleased changes are included in the returned version history, if
            available.

        Args:
            start_from (:class:`~chango.Version` | :obj:`str`, optional): The version to start
                from. If :obj:`None`, start from the earliest available version.
            end_at (:class:`~chango.Version` | :obj:`str`, optional): The version to end at.
                If :obj:`None`, end at the latest available version, *including* unreleased
                changes.

        Returns:
            :class:`VHT <typing.TypeVar>`: The loaded version :class:`~chango.abc.VersionHistory`.
        """
        version_history = self.build_version_history()

        if not end_at and self.scanner.has_unreleased_changes():
            version_history.add_version_note(self.load_version_note(None))

        for version in self.scanner.get_available_versions(start_from=start_from, end_at=end_at):
            version_history.add_version_note(self.load_version_note(version))

        return version_history

    def release(self, version: "Version") -> bool:
        """Release a version.
        This calls :meth:`get_write_directory` for all unreleased change notes and moves the file
        if necessary.

        Tip:
            This method calls :meth:`chango.abc.VersionScanner.invalidate_caches` after
            releasing the version.

        Args:
            version (:class:`~chango.Version`): The version to release.

        Returns:
            :obj:`bool`: Whether a release was performed. If no unreleased changes are available,
            this method returns :obj:`False`.
        """
        if not self.scanner.has_unreleased_changes():
            return False
        for uid in self.scanner.get_changes(None):
            change_info = self.scanner.lookup_change_note(uid)
            write_dir = self.get_write_directory(uid, version)
            if change_info.file_path.parent != write_dir:
                move_file(change_info.file_path, write_dir / change_info.file_path.name)

        self.scanner.invalidate_caches()

        return True



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import datetime as dtm
from unittest.mock import MagicMock

import pytest

from chango import Version
from chango.concrete import BackwardCompatibleVersionScanner
from chango.error import ChanGoError


class TestBackwardCompatibleVersionScanner:
    @staticmethod
    def build_mock_scanners(method_name: str, expected_results: list[object]) -> list[MagicMock]:
        mocks = [MagicMock() for _ in range(len(expected_results))]
        for mock, result in zip(mocks, expected_results, strict=False):
            method = getattr(mock, method_name)
            if isinstance(result, Exception):
                method.side_effect = result
            else:
                method.return_value = result
        return mocks

    @pytest.mark.parametrize(
        ("results", "expected"),
        [
            ([True, False], True),
            ([False, False], False),
            ([False, True], True),
            ([True, True], True),
        ],
    )
    @pytest.mark.parametrize("version", ["1.0.0", "2.0.0", None])
    def test_is_available(self, results, expected, version):
        scanners = self.build_mock_scanners("is_available", results)
        scanner = BackwardCompatibleVersionScanner(scanners)
        assert scanner.is_available(version) == expected

        was_true = False
        for scanner, result in zip(scanners, results, strict=False):
            if not was_true:
                scanner.is_available.assert_called_once_with(version)
                was_true = was_true or result
            else:
                assert not scanner.is_available.called

    @pytest.mark.parametrize(
        ("results", "expected"),
        [
            ([True, False], True),
            ([False, False], False),
            ([False, True], True),
            ([True, True], True),
        ],
    )
    def test_has_unreleased_changes(self, results, expected):
        scanners = self.build_mock_scanners("has_unreleased_changes", results)
        scanner = BackwardCompatibleVersionScanner(scanners)
        assert scanner.has_unreleased_changes() == expected

        was_true = False
        for scanner, result in zip(scanners, results, strict=False):
            if not was_true:
                scanner.has_unreleased_changes.assert_called_once_with()
                was_true = was_true or result
            else:
                assert not scanner.has_unreleased_changes.called

    @pytest.mark.parametrize(
        ("results", "expected"),
        [
            (
                [
                    Version("1.0.0", date=dtm.date(2025, 1, 1)),
                    Version("1.0.0", date=dtm.date(2025, 1, 10)),
                ],
                Version("1.0.0", date=dtm.date(2025, 1, 10)),
            ),
            (
                [
                    Version("1.0.0", date=dtm.date(2025, 1, 10)),
                    Version("1.0.0", date=dtm.date(2025, 1, 1)),
                ],
                Version("1.0.0", date=dtm.date(2025, 1, 10)),
            ),
            (
                [
                    ChanGoError("No versions available."),
                    Version("1.0.0", date=dtm.date(2025, 1, 10)),
                ],
                Version("1.0.0", date=dtm.date(2025, 1, 10)),
            ),
            (
                [ChanGoError("No versions available."), ChanGoError("No versions available.")],
                ChanGoError("No versions available."),
            ),
        ],
    )
    def test_get_latest_version(self, results, expected):
        scanners = self.build_mock_scanners("get_latest_version", results)
        scanner = BackwardCompatibleVersionScanner(scanners)
        if isinstance(expected, ChanGoError):
            with pytest.raises(ChanGoError, match=r"No versions available."):
                scanner.get_latest_version()
        else:
            assert scanner.get_latest_version() == expected

        for scanner in scanners:
            scanner.get_latest_version.assert_called_once_with()

    @pytest.mark.parametrize(
        ("results", "expected"),
        [
            (
                [
                    (Version("1.0.0", dtm.date.today()), Version("1.0.1", dtm.date.today())),
                    (Version("2.0.0", dtm.date.today()), Version("2.0.1", dtm.date.today())),
                ],
                (
                    Version("1.0.0", dtm.date.today()),
                    Version("1.0.1", dtm.date.today()),
                    Version("2.0.0", dtm.date.today()),
                    Version("2.0.1", dtm.date.today()),
                ),
            ),
            (
                [(Version("1.0.0", dtm.date.today()), Version("1.0.1", dtm.date.today())), []],
                (Version("1.0.0", dtm.date.today()), Version("1.0.1", dtm.date.today())),
            ),
        ],
    )
    def test_get_available_versions(self, results, expected):
        scanners = self.build_mock_scanners("get_available_versions", results)
        scanner = BackwardCompatibleVersionScanner(scanners)
        start_from = object()
        end_at = object()
        assert scanner.get_available_versions(start_from=start_from, end_at=end_at) == expected

        for scanner in scanners:
            scanner.get_available_versions.assert_called_once_with(start_from, end_at)

    @pytest.mark.parametrize(
        ("results", "expected"),
        [
            (
                [
                    ChanGoError("Change note '1.0.0' not available."),
                    ChanGoError("Change note '1.0.0' not available."),
                ],
                ChanGoError("Change note '1.0.0' not available."),
            ),
            (
                [
                    "ReturnValueA",
                    ChanGoError("Change note '1.0.0' not available."),
                    "ReturnValueB",
                ],
                "ReturnValueA",
            ),
            (
                [
                    ChanGoError("Change note '1.0.0' not available."),
                    "ReturnValueA",
                    "ReturnValueB",
                ],
                "ReturnValueA",
            ),
        ],
    )
    @pytest.mark.parametrize("method_name", ["lookup_change_note", "get_changes"])
    def test_lookup_change_note_get_changes(self, results, expected, method_name):
        scanners = self.build_mock_scanners(method_name, results)
        scanner = BackwardCompatibleVersionScanner(scanners)
        method_object = getattr(scanner, method_name)
        uid = object()
        if isinstance(expected, ChanGoError):
            with pytest.raises(ChanGoError, match=r"not available."):
                method_object(uid)
        else:
            assert method_object(uid) == expected

        has_returned = False
        for scanner, result in zip(scanners, results, strict=False):
            if not has_returned:
                getattr(scanner, method_name).assert_called_once_with(uid)
                has_returned = has_returned or not isinstance(result, ChanGoError)
            else:
                assert not getattr(scanner, method_name).called

    def test_invalidate_caches(self):
        scanners = self.build_mock_scanners("invalidate_caches", [None, None])
        scanner = BackwardCompatibleVersionScanner(scanners)
        scanner.invalidate_caches()
        for scanner in scanners:
            scanner.invalidate_caches.assert_called_once_with()



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import pytest
from pydantic import ValidationError

from chango.concrete.sections import PullRequest


class TestPullRequest:
    def test_init_required_args(self):
        pr = PullRequest(uid="uid1", author_uids=("author1",))
        assert pr.uid == "uid1"
        assert pr.author_uids == ("author1",)
        assert pr.closes_threads == ()

    def test_init_all_args(self):
        pr = PullRequest(uid="uid2", author_uids=("author2",), closes_threads=("thread3",))
        assert pr.uid == "uid2"
        assert pr.author_uids == ("author2",)
        assert pr.closes_threads == ("thread3",)

    def test_init_legacy_author_uid(self):
        pr = PullRequest(uid="uid3", author_uid="author3")
        assert pr.uid == "uid3"
        assert pr.author_uids == ("author3",)
        assert pr.closes_threads == ()

    def test_init_mutually_exclusive_author_fields(self):
        with pytest.raises(
            ValidationError, match="author_uid and author_uids are mutually exclusive"
        ):
            PullRequest(uid="uid4", author_uid="author4", author_uids=("author5",))

    def test_init_invalid_author_type(self):
        with pytest.raises(ValidationError, match="should be a valid tuple"):
            PullRequest(uid="uid5", author_uids=42)



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
from chango.concrete.sections import Section


class TestSection:
    def test_init_required_args(self):
        section = Section(uid="uid1", title="Title 1")
        assert section.uid == "uid1"
        assert section.title == "Title 1"
        assert section.is_required is False
        assert section.render_pr_details is True
        assert section.sort_order == 0

    def test_init_all_args(self):
        section = Section(
            uid="uid2", title="Title 2", is_required=True, render_pr_details=False, sort_order=1
        )
        assert section.uid == "uid2"
        assert section.title == "Title 2"
        assert section.is_required is True
        assert section.render_pr_details is False
        assert section.sort_order == 1



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
#
#  SPDX-License-Identifier: MIT
from collections.abc import Collection
from typing import Any, ClassVar, Self, override

from ...action import ChanGoActionData
from ._pullrequest import PullRequest
from ._sectionchangenote import SectionChangeNote


class GitHubSectionChangeNote(SectionChangeNote):
    """Specialization of :class:`~chango.concrete.sections.SectionChangeNote` for
    projects hosted on GitHub.

    Example:
        .. code-block:: python

            from chango.concrete.sections import GitHubSectionChangeNote, Section


            class MySectionChangeNote(
                GitHubSectionChangeNote.with_sections(
                    [
                        Section(uid="req_section", title="Required Section", is_required=True),
                        Section(uid="opt_section", title="Optional Section"),
                    ]
                )
            ):
                OWNER = "my-username"
                REPOSITORY = "my-repo"

    """

    OWNER: ClassVar[str | None] = None
    """:obj:`str`: The owner of the repository on GitHub. This must be set as a class variable."""
    REPOSITORY: ClassVar[str | None] = None
    """:obj:`str`: The name of the repository on GitHub. This must be set as a class variable."""

    @classmethod
    def _get_owner(cls) -> str:
        if cls.OWNER is None:
            raise ValueError("OWNER must be set as class variable.")
        return cls.OWNER

    @classmethod
    def _get_repository(cls) -> str:
        if cls.REPOSITORY is None:
            raise ValueError("REPOSITORY must be set as class variable.")
        return cls.REPOSITORY

    @classmethod
    @override
    def get_pull_request_url(cls, pr_uid: str) -> str:
        """Implementation of :meth:`SectionChangeNote.get_pull_request_url` based on
        :attr:`OWNER` and :attr:`REPOSITORY`.
        """
        return f"https://github.com/{cls._get_owner()}/{cls._get_repository()}/pull/{pr_uid}"

    @classmethod
    @override
    def get_thread_url(cls, thread_uid: str) -> str:
        """Implementation of :meth:`SectionChangeNote.get_pull_request_url` based on
        :attr:`OWNER` and :attr:`REPOSITORY`.
        """
        return f"https://github.com/{cls._get_owner()}/{cls._get_repository()}/issues/{thread_uid}"

    @classmethod
    @override
    def get_author_url(cls, author_uid: str) -> str:
        """Get the URL of an author with the given UID.

        Args:
            author_uid (:obj:`str`): The UID of an author as defined in
                :attr:`chango.concrete.sections.PullRequest.author_uids`.

        Returns:
            :obj:`str`: The URL of the author.

        """
        return f"https://github.com/{author_uid}"

    @classmethod
    def get_sections(
        cls,
        labels: Collection[str],  # noqa: ARG003
        issue_types: Collection[str] | None,  # noqa: ARG003
    ) -> set[str]:
        """Determine appropriate sections based on the labels of a pull request as well as
        the labels and types of the issues closed by the pull request.

        If this class has required sections, they are all returned.
        Otherwise, the first section in the order of
        :attr:`~chango.concrete.sections.Section.sort_order` is returned.

        Tip:
            This method can be overridden to provide custom logic for determining the
            section based on the labels and issue types.

        Args:
            labels (Collection[:obj:`str`]): The combined set of labels of the pull request and
                the issues closed by the pull request.
            issue_types (Collection[:obj:`str`]): The types of the issues closed by the pull
                request.

                Caution:
                    Since issue types are currently in
                    `public preview <https://github.com/orgs/community/discussions/139933>`_,
                    this set may be empty.

        Returns:
            Set[:obj:`str`]: The UIDs of the sections.

        """
        sorted_sections = sorted(
            (section for section in cls.SECTIONS.values()), key=lambda section: section.sort_order
        )
        required_sections = {section.uid for section in sorted_sections if section.is_required}
        return required_sections or {sorted_sections[0].uid}

    @classmethod
    @override
    def build_from_github_event(
        cls, event: dict[str, Any], data: dict[str, Any] | ChanGoActionData | None = None
    ) -> Self:
        """Implementation of :meth:`chango.abc.ChangeNote.build_from_github_event`.

        This writes the pull request title to the sections determined by :meth:`get_sections`.
        Uses the pull request number as slug.

        Caution:
            * Does not consider any formatting in the pull request title!
            * Considers the ``data`` argument only if it is an instance of
              :class:`~chango.action.ChanGoActionData`.

        Raises:
            ValueError: If required data is missing or not in the expected format.
        """
        try:
            pull_request = event["pull_request"]
            pr_number = pull_request["number"]
            pr_title = pull_request["title"]
            pr_labels = {label["name"] for label in pull_request.get("labels", [])}
            author_uid = pull_request["user"]["login"]
        except (KeyError, TypeError) as exc:
            raise ValueError("Unable to extract required data from event.") from exc

        issue_types: set[str] = set()
        closes_threads: set[int] = set()
        labels = pr_labels

        if isinstance(data, ChanGoActionData) and data.linked_issues:
            for issue in data.linked_issues:
                closes_threads.add(issue.number)
                if issue.labels:
                    labels.update(issue.labels)
                if issue.issue_type:
                    issue_types.add(issue.issue_type)

        sections = cls.get_sections(labels, issue_types)
        return cls(
            slug=f"{pr_number:04}",  # type: ignore[call-arg]
            pull_requests=(
                PullRequest(
                    uid=str(pr_number),
                    author_uids=(author_uid,),
                    closes_threads=tuple(map(str, closes_threads)),
                ),
            ),
            **dict.fromkeys(sections, pr_title),
        )



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT

from pathlib import Path
from unittest.mock import MagicMock

from tests.cli.conftest import ReuseCliRunner


class TestEdit:
    def test_release_no_unreleased(self, cli: ReuseCliRunner, mock_chango_instance, monkeypatch):
        launch_mock = MagicMock()
        monkeypatch.setattr("typer.launch", launch_mock)

        test_path = Path("this/is/a/test/path")
        mock_chango_instance.scanner.lookup_change_note.return_value.file_path = test_path

        result = cli.invoke(args=["edit", "some_uid"])

        assert result.check_exit_code()
        assert result.stdout == ""

        mock_chango_instance.scanner.lookup_change_note.assert_called_once_with("some_uid")
        launch_mock.assert_called_once_with(test_path.as_posix())



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
#
#  SPDX-License-Identifier: MIT
import abc
import tomllib
from collections.abc import Collection
from typing import Any, ClassVar, Final, Self, override

import pydantic as pydt
import tomlkit

from chango._utils.files import UTF8
from chango.abc import ChangeNote
from chango.concrete.sections._pullrequest import PullRequest
from chango.concrete.sections._section import Section
from chango.constants import MarkupLanguage
from chango.error import ValidationError


class SectionChangeNote(pydt.BaseModel, ChangeNote, abc.ABC):
    """A change note that consists of multiple sections and includes references to pull requests
    that are related to the change.

    Uses the `toml <https://toml.io/>`_ format for specifying the content.

    Important:
        * This class does not contain any specified sections by default and must not be
          instantiated directly. Use :meth:`with_sections` to create a suitable subclass with the
          desired sections.
        * Even though this class is in the :mod:`~chango.concrete` module, it is still an abstract
          base class and must be subclassed to be used. However, only the methods
          :meth:`get_pull_request_url`, :meth:`get_thread_url`, and :meth:`get_author_url` must be
          implemented in the subclass. A concrete subclass is provided in
          :class:`~chango.concrete.sections.GitHubSectionChangeNote`.

    Args:
        pull_requests (tuple[:class:`PullRequest`], optional): The pull requests that are related
            to the change.

    Attributes:
        pull_requests (tuple[:class:`~chango.concrete.sections.PullRequest`]): The pull
            requests that are related to the change
    """

    MARKUP: ClassVar[str] = MarkupLanguage.RESTRUCTUREDTEXT
    """:obj:`str`: The markup language used in the sections.
    """

    SECTIONS: Final[dict[str, Section]] = pydt.Field(default_factory=dict)
    """dict[:obj:`str`, :class:`Section`]: The sections of the change note. Maps the UID of the
    section to the :class:`Section` object containing the configuration for the section
    """

    _BUILT_BY_WITH_SECTIONS: Final[bool] = pydt.PrivateAttr(default=False)
    """Internal attribute to check if the class was built by the :meth:`with_sections`
    class method.
    """

    pull_requests: tuple[PullRequest, ...] = pydt.Field(default_factory=tuple)

    def __init__(self, slug: str, *args: Any, uid: str | None = None, **kwargs: Any) -> None:
        # Mixing pydantic with non-pydantic base classes is a bit tricky.
        # Unfortunately, we have to call the __init__ methods of both classes manually and also
        # don't get an overly nice signature. However, this class should rarely be instantiated
        # directly, so this should be acceptable.
        super().__init__(*args, **kwargs)
        ChangeNote.__init__(self, slug=slug, uid=uid)

        if not any(getattr(self, name) for name in self.SECTIONS):
            raise ValidationError("At least one section must be specified")

    @pydt.model_validator(mode="after")
    def _validate_section_configuration(self) -> Self:
        # Runs after the pydantic validation. Adds some additional checks in case someone is
        # trying to manually build subclasses of SectionChangeNote.
        if not self._BUILT_BY_WITH_SECTIONS:
            raise TypeError(
                "SectionChangeNote must not be subclassed manually. Please use"
                "SectionChangeNote.with_sections to create a suitable subclass."
            )

        return self

    @classmethod
    def with_sections(cls, sections: Collection[Section], name: str | None = None) -> type[Self]:
        """Create a new subclass of :class:`SectionChangeNote` with the given sections.

        Args:
            sections (Collection[:class:`Section`]): The sections to include in the
                change note.

                Tip:
                    All sections may be optional, but at least one section must be specified
                    on instantiation. That is, a change note without content in any section is
                    not allowed.
            name (:obj:`str`, optional): The name of the new class. Defaults to
                ``DynamicSectionChangeNote``.

        Returns:
            type[:class:`SectionChangeNote`]: The new subclass of :class:`SectionChangeNote`.

        """
        # This also covers the case of `sections` being an empty collection
        if not sections:
            raise ValueError("Class must have at least one section")

        config_fields = {
            section.uid: (str, pydt.Field(...)) if section.is_required else (str | None, None)
            for section in sections
        }

        doc_insert = ", ".join(f'"{x}"' for x in (section.title for section in sections))

        dynamic_model = pydt.create_model(  # type: ignore[call-overload]
            name or "DynamicSectionChangeNote",
            __base__=cls,
            __doc__=f"SectionChangeNote with sections {doc_insert}",
            **config_fields,
        )
        dynamic_model.SECTIONS = {section.uid: section for section in sections}

        dynamic_model._BUILT_BY_WITH_SECTIONS = True
        return dynamic_model

    @property
    @override
    def file_extension(self) -> str:
        return "toml"

    @classmethod
    @override
    def from_string(cls, slug: str, uid: str, string: str) -> Self:
        """Implementation of :meth:`~chango.abc.ChangeNote.from_string`.

        Args:
            slug (:obj:`str`): The slug of the change note.
            uid (:obj:`str`): The UID of the change note.
            string (:obj:`str`): The ``toml`` string to read from.

        Returns:
            :class:`~chango.abc.ChangeNote`: The :class:`~chango.abc.ChangeNote` object.

        Raises:
            :class:`chango.error.ValidationError`: If the string is not a valid change note.

        """
        try:
            return cls(slug=slug, uid=uid, **tomllib.loads(string))
        except (tomllib.TOMLDecodeError, pydt.ValidationError) as exc:
            raise ValidationError(f"Invalid TOML data: {exc}") from exc

    @override
    def to_string(self, encoding: str = UTF8) -> str:
        return tomlkit.dumps(
            {key: value for key, value in self.model_dump().items() if value is not None}
        )

    @classmethod
    @override
    def build_template(cls, slug: str, uid: str | None = None) -> Self:
        required_sections = {
            field_name: "Required Section Content"
            if field_info.is_required()
            else "Optional Section Content"
            for field_name, field_info in cls.model_fields.items()
            if field_name != "pull_requests"
        }
        pull_requests = (
            PullRequest(
                uid="pr-number-1", closes_threads=("thread1", "thread2"), author_uids=("author1",)
            ),
            PullRequest(uid="pr-number-2", closes_threads=("thread3",), author_uids=("author2",)),
        )
        return cls(slug=slug, uid=uid, pull_requests=pull_requests, **required_sections)

    @classmethod
    @abc.abstractmethod
    def get_pull_request_url(cls, pr_uid: str) -> str:
        """Get the URL of the pull request with the given UID.

        Args:
            pr_uid (:obj:`str`): The UID of the pull request as defined in
                :attr:`chango.concrete.sections.PullRequest.uid`.

        Returns:
            :obj:`str`: The URL of the pull request.

        """

    @classmethod
    @abc.abstractmethod
    def get_thread_url(cls, thread_uid: str) -> str:
        """Get the URL of the thread with the given UID.

        Args:
            thread_uid (:obj:`str`): The UID of the thread as defined in
                :attr:`chango.concrete.sections.PullRequest.closes_threads`.

        Returns:
            :obj:`str`: The URL of the thread.

        """

    @classmethod
    @abc.abstractmethod
    def get_author_url(cls, author_uid: str) -> str:
        """Get the URL of an author with the given UID.

        Args:
            author_uid (:obj:`str`): The UID of an author as defined in
                :attr:`chango.concrete.sections.PullRequest.author_uids`.

        Returns:
            :obj:`str`: The URL of the author.

        """



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT

from pathlib import Path
from unittest.mock import MagicMock

import pytest

from tests.cli.conftest import ReuseCliRunner


class TestNew:
    @pytest.mark.parametrize("slug_option", ["--slug", "-s"], ids=["LongSlug", "ShortSlug"])
    @pytest.mark.parametrize(
        ("edit", "edit_option"),
        [(True, "--edit"), (True, "-e"), (False, "--no-edit"), (False, "-ne"), (None, None)],
    )
    def test_new(
        self,
        cli: ReuseCliRunner,
        mock_chango_instance,
        monkeypatch,
        edit,
        slug_option,
        edit_option,
    ):
        launch_mock = MagicMock()
        monkeypatch.setattr("typer.launch", launch_mock)

        test_path = Path("this/is/a/test/path")
        change_note = mock_chango_instance.build_template_change_note.return_value
        change_note.file_name = "expected_file_name"
        mock_chango_instance.write_change_note.return_value = test_path

        args = ["new", slug_option, "some_uid"]
        if edit_option is not None:
            args.append(edit_option)
        result = cli.invoke(args=args)

        assert result.check_exit_code()
        assert result.stdout == "Created new change note expected_file_name\n"

        mock_chango_instance.build_template_change_note.assert_called_once_with(slug="some_uid")
        mock_chango_instance.write_change_note.assert_called_once_with(change_note, version=None)
        if edit in (True, None):
            launch_mock.assert_called_once_with(test_path.as_posix())
        else:
            launch_mock.assert_not_called()



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import abc
import warnings
from collections.abc import Iterator, MutableMapping

from .._utils.types import VersionUID, VUIDInput
from ..abc._versionnote import VersionNote
from ..helpers import ensure_uid


class VersionHistory[VNT: VersionNote](MutableMapping[VersionUID, VNT], abc.ABC):
    """Abstract base class for a version history describing the versions in a software project over
    several versions.

    Hint:
        Objects of this class can be used as :class:`~collections.abc.MutableMapping`, where the
        keys are the unique identifiers of the versions and the values are the version notes
        themselves.
    """

    def __init__(self) -> None:
        self._version_notes: dict[VersionUID, VNT] = {}

    def __delitem__(self, key: VUIDInput, /) -> None:
        del self._version_notes[ensure_uid(key)]

    def __getitem__(self, key: VUIDInput, /) -> VNT:
        return self._version_notes[ensure_uid(key)]

    def __iter__(self) -> Iterator[VersionUID]:
        return iter(self._version_notes)

    def __len__(self) -> int:
        return len(self._version_notes)

    def __setitem__(self, key: VUIDInput, value: VNT, /) -> None:
        if ensure_uid(key) != value.uid:
            warnings.warn(
                f"Key {key!r} does not match version note UID {value.uid!r}. "
                "Using the version UID as key.",
                stacklevel=2,
            )
        self._version_notes[value.uid] = value  # type: ignore[index]

    def add_version_note(self, version_note: VNT) -> None:
        """Add a version note to the version note.

        Args:
            version_note (:class:`VNT <typing.TypeVar>`): The :class:`~chango.abc.VersionNote`
                note to add.
        """
        self[version_note.uid] = version_note  # type: ignore[index]

    def remove_version_note(self, version_note: VNT) -> None:
        """Remove a version note from the version note.

        Args:
            version_note (:class:`VNT <typing.TypeVar>`): The :class:`~chango.abc.VersionNote`
                note to remove.
        """
        del self[version_note.uid]  # type: ignore[arg-type]

    @abc.abstractmethod
    def render(self, markup: str) -> str:
        """Render the version note as a string. Must include information about all change notes
        contained in the version note.

        Hint:
            * Make use of :meth:`chango.abc.VersionNote.render` to render the change notes.
            * The change notes should be rendered in reverse chronological order. This needs to be
              handled by the implementation and can be achieved either by applying appropriate
              sorting the :attr:`~chango.abc.VersionNote.uid` or by sorting by
              :attr:`~chango.abc.VersionNote.date` if available.

        Args:
            markup (:obj:`str`): The markup language to use for rendering. If the markup language
                is not supported, an :exc:`~chango.error.UnsupportedMarkupError` should be raised.

        Returns:
            :obj:`str`: The rendered version note.
        """



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import datetime as dtm

import pytest

from chango import Version
from chango.concrete import CommentChangeNote, CommentVersionNote
from chango.error import UnsupportedMarkupError


class TestCommentVersionNote:
    comments = ("comment 1", "a\nmulti-line\ncomment 2", "comment 3")
    change_notes = tuple(
        CommentChangeNote(slug=f"slug-{i}", comment=comment) for i, comment in enumerate(comments)
    )

    @pytest.fixture(
        autouse=True,
        params=[None, Version(uid="1.0.0", date=dtm.date(2024, 1, 1))],
        ids=["None", "Version"],
    )
    def setup(self, request):
        # This is the next best thing to parametrizing __init__ that I could find
        # in reasonable time
        version = request.param
        self.version = version
        self.version_note = CommentVersionNote(version=version)
        for change_note in self.change_notes:
            self.version_note.add_change_note(change_note)

    def test_unsupported_markup(self):
        with pytest.raises(UnsupportedMarkupError):
            self.version_note.render("unsupported markup")

    def test_markdown(self):
        rendered = self.version_note.render("markdown")
        assert (
            rendered
            == """- comment 1
- a

    multi-line

    comment 2
- comment 3"""
        )

    def test_html(self):
        rendered = self.version_note.render("html")
        assert (
            rendered
            == """<ul>
<li>comment 1</li>
<li>a<br>multi-line<br>comment 2</li>
<li>comment 3</li>
</ul>"""
        )

    def test_restructuredtext(self):
        rendered = self.version_note.render("restructuredtext")
        assert (
            rendered
            == """- comment 1
- a

  multi-line

  comment 2
- comment 3"""
        )

    def test_fallback(self):
        rendered = self.version_note.render("text")
        assert (
            rendered
            == """comment 1

a
multi-line
comment 2

comment 3"""
        )



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
from pathlib import Path
from types import SimpleNamespace

import pytest

from chango._utils.types import PathLike
from chango.config import ChanGoInstanceConfig, get_chango_instance
from tests.auxil.files import TEST_DATA_PATH

# INFO:
# Best reference for how use sphinx testing so far is
# https://github.com/sphinx-doc/sphinx/issues/7008


@pytest.fixture(scope="session")
def rootdir():
    return TEST_DATA_PATH / "sphinx_ext"


class MockStorage:
    def __init__(self):
        self.loaded_config: MockCGConfig | None = None

    def invalidate_storage(self):
        self.loaded_config = None

    def set_current_config(self, config: "MockCGConfig"):
        self.loaded_config = config

    def get(self) -> "MockCGConfig":
        if self.loaded_config is None:
            raise RuntimeError("No config loaded")
        return self.loaded_config

    @property
    def rendered_content(self):
        return self.loaded_config.chango.version_history.RENDERED_CONTENT


CG_CONFIG_STORAGE = MockStorage()


class MockVersionHistory(SimpleNamespace):
    RENDERED_CONTENT = "This is the rendered version history"

    def __init__(self):
        super().__init__()
        self.received_kwargs = None
        self.received_args = None

    def render(self, *args, **kwargs) -> str:
        self.received_args = args
        self.received_kwargs = kwargs
        return self.RENDERED_CONTENT


class MockChanGo(SimpleNamespace):
    def __init__(self):
        super().__init__()
        self.received_kwargs = None
        self.received_args = None
        self.version_history = MockVersionHistory()

    def load_version_history(
        self, *args, start_from: str | None = None, end_at: str | None = None, **kwargs
    ) -> MockVersionHistory:
        self.received_args = args
        self.received_kwargs = kwargs | {"start_from": start_from, "end_at": end_at}
        return self.version_history


class MockCGConfig(SimpleNamespace):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.chango = MockChanGo()

    @classmethod
    def load(cls, path: PathLike | None):
        out = cls(
            sys_path=None if path is None else Path(path),
            chango_instance=ChanGoInstanceConfig(name="name", module="module"),
        )
        CG_CONFIG_STORAGE.set_current_config(out)
        return out

    def import_chango_instance(self) -> MockChanGo:
        return self.chango


@pytest.fixture(autouse=True)
def cg_config_mock(monkeypatch):
    monkeypatch.setattr("chango.config.ChanGoConfig", MockCGConfig)
    yield CG_CONFIG_STORAGE
    get_chango_instance.cache_clear()
    CG_CONFIG_STORAGE.invalidate_storage()



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
from collections import defaultdict
from typing import TYPE_CHECKING, override

from ..._utils.strings import indent_multiline
from ...abc import VersionNote
from ...constants import MarkupLanguage
from ...error import UnsupportedMarkupError
from ._pullrequest import PullRequest
from ._sectionchangenote import SectionChangeNote

if TYPE_CHECKING:
    from chango import Version


class SectionVersionNote[V: (Version, None), SCN: SectionChangeNote](VersionNote[SCN, V]):
    """An implementation of :class:`~chango.abc.VersionNote` that works with
    :class:`~chango.concrete.sections.SectionChangeNote`.

    Important:
        Currently, only :attr:`~chango.constants.MarkupLanguage.RESTRUCTUREDTEXT` is supported.

    Args:
        section_change_note_type (\
            type[:class:`~chango.concrete.sections.SectionChangeNote`]): The type of
            the section change note to use.

            Hint:
                It will not be possible to add change notes of a different type to this version
                note.
    """

    def __init__(self, version: V, section_change_note_type: type[SCN]) -> None:
        super().__init__(version)  # type: ignore[arg-type]

        if section_change_note_type.MARKUP != MarkupLanguage.RESTRUCTUREDTEXT:
            raise UnsupportedMarkupError(
                "This version note currently only supports reStructuredText markup."
            )

        self._section_change_note_type = section_change_note_type
        self._sorted_sections = dict(
            sorted(section_change_note_type.SECTIONS.items(), key=lambda x: x[1].sort_order)
        )

    @override
    def __setitem__(self, key: str, value: SCN, /) -> None:
        if not isinstance(value, self._section_change_note_type):
            raise TypeError(
                f"Expected a {self._section_change_note_type} instance, got {type(value)}"
            )
        super().__setitem__(key, value)

    def _render_pr(self, pr: PullRequest) -> str:
        pr_url = self._section_change_note_type.get_pull_request_url(pr.uid)

        author_links = [
            f"`@{author_uid} <{self._section_change_note_type.get_author_url(author_uid)}>`_"
            for author_uid in pr.author_uids
        ]

        thread_links = [
            f"`#{thread_uid} <{self._section_change_note_type.get_thread_url(thread_uid)}>`_"
            for thread_uid in pr.closes_threads
        ]

        base = f"`#{pr.uid} <{pr_url}>`_ by {', '.join(author_links)}"
        if not thread_links:
            return base
        return f"{base} closes {', '.join(thread_links)}"

    def _render_section_entry(
        self, content: str, pull_requests: tuple[PullRequest, ...] | None = None
    ) -> str:
        indented_content = f"- {indent_multiline(content, newlines=2)}"

        if not pull_requests:
            return indented_content

        pr_details = "; ".join(self._render_pr(pr) for pr in pull_requests)
        if "\n" not in content:
            return f"{indented_content} ({pr_details})"
        return f"{indented_content}\n\n  ({pr_details})"

    @override
    def render(self, markup: str) -> str:
        """Render the version note to a string by listing all contained change notes.
        Aggregates the content of all change notes for each section and renders them in the order
        defined by :attr:`~chango.concrete.sections.Section.sort_order`.

        Important:
            Currently, only :attr:`~chango.constants.MarkupLanguage.RESTRUCTUREDTEXT` is supported.
        """
        try:
            markup = MarkupLanguage.from_string(markup)
        except ValueError as exc:
            raise UnsupportedMarkupError(markup) from exc

        if markup != MarkupLanguage.RESTRUCTUREDTEXT:
            raise UnsupportedMarkupError(markup)

        section_contents: dict[str, str] = defaultdict(str)
        for change_note in self.values():
            for section_uid, section in self._sorted_sections.items():
                if section_content := getattr(change_note, section_uid):
                    section_contents[section_uid] = "\n".join(
                        (
                            section_contents[section_uid],
                            self._render_section_entry(
                                section_content,
                                change_note.pull_requests if section.render_pr_details else None,
                            ),
                        )
                    )

        return "\n\n".join(
            f"{section.title}\n{'-' * len(section.title)}\n{content}"
            for uid, section in self._sorted_sections.items()
            if (content := section_contents[uid])
        )



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
from typing import Annotated, Any

import pydantic as pydt
from pydantic import BeforeValidator


def _validate_author_uid(value: str | tuple[str, ...]) -> tuple[str, ...]:
    if isinstance(value, str):
        return (value,)
    return value


class PullRequest(pydt.BaseModel):
    """Simple data structure to represent a pull/merge request.

    Args:
        uid (:obj:`str`): The unique identifier for the pull request. For example, the pull request
            number.
        author_uids (:obj:`str` | tuple[:obj:`str`, ...]): The unique identifier of the author(s)
            of the pull request. For example, the author's username.
        closes_threads (tuple[:obj:`str`], optional): The threads that are closed by this pull
            request.

    Attributes:
        uid (:obj:`str`): The unique identifier for the pull request.
        author_uids (tuple[:obj:`str`, ...]): The unique identifier of the author(s) of the pull
            request.
        closes_threads (tuple[:obj:`str`]): The threads that are closed by this pull request.
            May be empty.

    """

    uid: str
    author_uids: Annotated[tuple[str, ...], BeforeValidator(_validate_author_uid)]
    closes_threads: tuple[str, ...] = pydt.Field(default_factory=tuple)

    @pydt.model_validator(mode="before")
    def unify_author_ids(cls, data: Any) -> Any:
        # for backwards compatibility, we allow both `author_uid` and `author_uids`
        if not isinstance(data, dict):
            # This cause is here only due to pydantics documentation example.
            # in practice, this should never happen.
            return data  # pragma: no cover

        author_uid = data.pop("author_uid", None)
        author_uids = data.pop("author_uids", None)

        if author_uid is not None and author_uids is not None:
            raise ValueError("author_uid and author_uids are mutually exclusive")
        data["author_uids"] = author_uid or author_uids
        return data



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import os
from contextlib import contextmanager
from pathlib import Path
from types import SimpleNamespace

from chango.config import ChanGoConfig, get_chango_instance


@contextmanager
def temporary_chdir(path: Path):
    current_dir = Path.cwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(current_dir)


class TestConfigModule:
    def test_get_chango_instance_caching(self, monkeypatch):
        call_count = 0

        def import_chango_instance():
            nonlocal call_count
            call_count += 1
            return call_count

        def load(*_, **__):
            return SimpleNamespace(import_chango_instance=import_chango_instance)

        monkeypatch.setattr(ChanGoConfig, "load", load)

        for _ in range(10):
            assert get_chango_instance() == 1, "The ChanGo instance should be cached!"

        # to ensure that other tests still pass
        get_chango_instance.cache_clear()



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
from chango.__about__ import __version__
from tests.cli.conftest import ReuseCliRunner


class TestMain:
    def test_version(self, cli: ReuseCliRunner):
        result = cli.invoke(args=["--version"])
        assert result.check_exit_code()
        assert result.stdout == __version__ + "\n"



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
from typing import Any, ClassVar, Self, override

from .._utils.files import UTF8
from ..abc import ChangeNote
from ..action import ChanGoActionData
from ..constants import MarkupLanguage


class CommentChangeNote(ChangeNote):
    """A simple change note that consists of a single comment. May be multi-line.

    Args:
        comment (:obj:`str`): The comment text.

    Attributes:
        comment (:obj:`str`): The comment text.
    """

    MARKUP: ClassVar[str] = MarkupLanguage.TEXT
    """:obj:`str`: The markup language used in the comment. Will also be used as file extension.
    """

    @override
    def __init__(self, slug: str, comment: str, uid: str | None = None):
        super().__init__(slug=slug, uid=uid)
        self.comment: str = comment

    @property
    @override
    def file_extension(self) -> str:
        return self.MARKUP

    @classmethod
    @override
    def from_string(cls, slug: str, uid: str, string: str) -> Self:
        return cls(slug=slug, comment=string, uid=uid)

    @override
    def to_string(self, encoding: str = UTF8) -> str:
        return self.comment

    @classmethod
    @override
    def build_template(cls, slug: str, uid: str | None = None) -> Self:
        return cls(slug=slug, comment="example comment", uid=uid)

    @classmethod
    @override
    def build_from_github_event(
        cls, event: dict[str, Any], data: dict[str, Any] | ChanGoActionData | None = None
    ) -> Self:
        """Implementation of :meth:`~chango.abc.ChangeNote.build_from_github_event`.

        Considers only events of type ``pull_request`` and ``pull_request_target``.
        Uses the pull request number as slug and the pull request title as comment.

        Currently only supports :attr:`~chango.constants.MarkupLanguage.TEXT`,
            :attr:`~chango.constants.MarkupLanguage.MARKDOWN`,
            :attr:`~chango.constants.MarkupLanguage.RESTRUCTUREDTEXT` and
            :attr:`~chango.constants.MarkupLanguage.HTML`.

        Caution:
            Does not consider any formatting in the pull request title!

        Raises:
            ValueError: If required data is missing or not in the expected format or
                if :attr:`MARKUP` is not supported.
        """
        try:
            pull_request = event["pull_request"]
            pr_number = pull_request["number"]
            html_url = pull_request["html_url"]
        except KeyError as exc:
            raise ValueError("Unable to extract required data from event.") from exc

        match cls.MARKUP:
            case MarkupLanguage.TEXT:
                link = f"({html_url})"
            case MarkupLanguage.MARKDOWN:
                link = f"([#{pr_number}]({html_url}))"
            case MarkupLanguage.RESTRUCTUREDTEXT:
                link = f"(`#{pr_number} <{html_url}>`_)"
            case MarkupLanguage.HTML:
                link = f'(<a href="{html_url}">#{pr_number}</a>)'
            case _:
                raise ValueError(f"Unsupported markup language: {cls.MARKUP}")

        return cls(slug=f"{pr_number:04}", comment=f"{pull_request['title']} {link}")



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import datetime as dtm

import pytest

from chango import Version
from chango.abc import VersionScanner
from chango.concrete import DirectoryVersionScanner
from chango.error import ChanGoError
from tests.auxil.files import data_path


@pytest.fixture
def scanner(monkeypatch) -> DirectoryVersionScanner:
    # DVS overrides get_version, but we want to test the base implementation
    monkeypatch.setattr(DirectoryVersionScanner, "get_version", VersionScanner.get_version)
    return DirectoryVersionScanner(TestVersionScanner.DATA_ROOT, "unreleased")


@pytest.fixture
def scanner_no_unreleased(monkeypatch) -> DirectoryVersionScanner:
    # DVS overrides get_version, but we want to test the base implementation
    monkeypatch.setattr(DirectoryVersionScanner, "get_version", VersionScanner.get_version)
    return DirectoryVersionScanner(TestVersionScanner.DATA_ROOT, "no-unreleased")


class TestVersionScanner:
    """Since VersionScanner is an abstract base class, we are testing with DirectoryVersionScanner
    as a simple implementation.

    Note that we do *not* test abstract methods, as that is the responsibility of the concrete
    implementations.
    """

    DATA_ROOT = data_path("directoryversionscanner")

    @pytest.mark.parametrize(
        ("version", "expected"),
        [
            ("1.1", True),
            (Version("1.1", dtm.date(2024, 1, 1)), True),
            (Version("1.1", dtm.date(2024, 5, 1)), False),
            ("1.2", True),
            (Version("1.2", dtm.date(2024, 1, 2)), True),
            (Version("1.2", dtm.date(2024, 5, 1)), False),
            ("1.3", True),
            (Version("1.3", dtm.date(2024, 1, 3)), True),
            (Version("1.3", dtm.date(2024, 5, 1)), False),
            ("1.3.1", True),
            (Version("1.3.1", dtm.date(2024, 1, 3)), True),
            (Version("1.3.1", dtm.date(2024, 5, 1)), False),
            (None, True),
            ("1.4", False),
            ("1.0", False),
            (object(), False),
            (dtm.date(2024, 1, 1), False),
        ],
    )
    def test_contains(self, scanner, version, expected):
        assert (version in scanner) == expected

    def test_contains_no_unreleased_changes(self, scanner_no_unreleased):
        assert None not in scanner_no_unreleased

    def test_iter(self, scanner):
        assert set(scanner) == {
            Version("1.1", dtm.date(2024, 1, 1)),
            Version("1.2", dtm.date(2024, 1, 2)),
            Version("1.3", dtm.date(2024, 1, 3)),
            Version("1.3.1", dtm.date(2024, 1, 3)),
        }

    def test_len(self, scanner):
        assert len(scanner) == len(scanner.get_available_versions())

    @pytest.mark.parametrize("idx", [1, 2, 3])
    def test_get_version(self, scanner, idx):
        version = scanner.get_version(f"1.{idx}")
        assert version.uid == f"1.{idx}"
        assert version.date == dtm.date(2024, 1, idx)

    def test_get_version_not_found(self, scanner):
        with pytest.raises(ChanGoError, match="not available"):
            scanner.get_version("1.4")

    def test_invalidates_caches(self, scanner):
        # This does nothing, but we want to test that it doesn't raise an error
        scanner.invalidate_caches = VersionScanner.invalidate_caches
        scanner.invalidate_caches(scanner)



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT

import datetime as dtm
from pathlib import Path

import pytest

from chango import Version
from chango.concrete import DirectoryVersionScanner
from chango.error import ChanGoError
from tests.auxil.files import data_path


@pytest.fixture
def scanner() -> DirectoryVersionScanner:
    return DirectoryVersionScanner(TestDirectoryVersionScanner.DATA_ROOT, "unreleased")


@pytest.fixture
def scanner_no_unreleased() -> DirectoryVersionScanner:
    return DirectoryVersionScanner(TestDirectoryVersionScanner.DATA_ROOT, "no-unreleased")


class TestDirectoryVersionScanner:
    DATA_ROOT = data_path("directoryversionscanner")

    def test_init_basic(self):
        scanner = DirectoryVersionScanner(self.DATA_ROOT, "unreleased")
        assert scanner.base_directory == self.DATA_ROOT
        assert scanner.unreleased_directory == self.DATA_ROOT / "unreleased"

        scanner = DirectoryVersionScanner(self.DATA_ROOT, self.DATA_ROOT / "unreleased")
        assert scanner.base_directory == self.DATA_ROOT
        assert scanner.unreleased_directory == self.DATA_ROOT / "unreleased"

    def test_init_relative(self):
        scanner = DirectoryVersionScanner("../data/directoryversionscanner", "unreleased")
        assert scanner.base_directory == self.DATA_ROOT
        assert scanner.unreleased_directory == self.DATA_ROOT / "unreleased"

    def test_base_directory_not_exists(self):
        with pytest.raises(ValueError, match="does not exist"):
            DirectoryVersionScanner("does_not_exist", "unreleased")

    def test_unreleased_directory_not_exists(self):
        with pytest.raises(ValueError, match="does not exist"):
            DirectoryVersionScanner(self.DATA_ROOT, "does_not_exist")

    @pytest.mark.parametrize(
        ("version", "expected"),
        [
            ("1.1", True),
            (Version("1.1", dtm.date(2024, 1, 1)), True),
            (Version("1.1", dtm.date(2024, 5, 1)), False),
            ("1.2", True),
            (Version("1.2", dtm.date(2024, 1, 2)), True),
            (Version("1.2", dtm.date(2024, 5, 1)), False),
            ("1.3", True),
            (Version("1.3", dtm.date(2024, 1, 3)), True),
            (Version("1.3", dtm.date(2024, 5, 1)), False),
            ("1.3.1", True),
            (Version("1.3.1", dtm.date(2024, 1, 3)), True),
            (Version("1.3.1", dtm.date(2024, 5, 1)), False),
            (None, True),
            ("1.4", False),
            ("1.0", False),
        ],
    )
    def test_is_available(self, scanner, version, expected):
        assert scanner.is_available(version) == expected

    def test_is_available_no_unreleased(self, scanner_no_unreleased):
        assert scanner_no_unreleased.is_available(None) is False

    def test_has_unreleased_changes(self, scanner, scanner_no_unreleased):
        assert scanner.has_unreleased_changes() is True
        assert scanner_no_unreleased.has_unreleased_changes() is False

    def test_get_latest_version(self, scanner):
        # It's important that we get 1.3.1 here and not 1.3, since both are released
        # on the same day such that lexicographical sorting kicks in!
        latest = scanner.get_latest_version()
        assert latest.uid == "1.3.1"
        assert latest.date == dtm.date(2024, 1, 3)

    def test_get_latest_version_nothing_released(self):
        scanner = DirectoryVersionScanner(self.DATA_ROOT / "no-released", "unreleased")
        with pytest.raises(ChanGoError, match="No versions available"):
            scanner.get_latest_version()

    def test_get_available_versions(self, scanner):
        assert set(scanner.get_available_versions()) == {
            Version("1.1", dtm.date(2024, 1, 1)),
            Version("1.2", dtm.date(2024, 1, 2)),
            Version("1.3", dtm.date(2024, 1, 3)),
            Version("1.3.1", dtm.date(2024, 1, 3)),
        }

    def test_get_available_versions_start_from(self, scanner):
        assert set(scanner.get_available_versions(start_from="1.2")) == {
            Version("1.2", dtm.date(2024, 1, 2)),
            Version("1.3", dtm.date(2024, 1, 3)),
            Version("1.3.1", dtm.date(2024, 1, 3)),
        }

    def test_get_available_versions_end_at(self, scanner):
        assert set(scanner.get_available_versions(end_at="1.2")) == {
            Version("1.1", dtm.date(2024, 1, 1)),
            Version("1.2", dtm.date(2024, 1, 2)),
        }

    @pytest.mark.parametrize("idx", [1, 2, 3])
    def test_lookup_change_note(self, scanner, idx):
        change_note = scanner.lookup_change_note(f"uid_1-{idx}_0")
        assert change_note.uid == f"uid_1-{idx}_0"
        assert change_note.version == Version(f"1.{idx}", dtm.date(2024, 1, idx))
        assert (
            change_note.file_path
            == self.DATA_ROOT
            / f"1.{idx}_2024-01-0{idx}"
            / f"comment-change-note.uid_1-{idx}_0.txt"
        )

    def test_lookup_change_note_unreleased(self, scanner):
        change_note = scanner.lookup_change_note("uid_ur_0")
        assert change_note.uid == "uid_ur_0"
        assert change_note.version is None
        assert (
            change_note.file_path
            == self.DATA_ROOT / "unreleased" / "comment-change-note.uid_ur_0.txt"
        )

    def test_lookup_change_note_not_found(self, scanner):
        with pytest.raises(ChanGoError, match="not found in any version"):
            scanner.lookup_change_note("unknown_uid")

    @pytest.mark.parametrize("idx", [1, 2, 3])
    def test_get_version(self, scanner, idx):
        version = scanner.get_version(f"1.{idx}")
        assert version.uid == f"1.{idx}"
        assert version.date == dtm.date(2024, 1, idx)

    def test_get_version_not_found(self, scanner):
        with pytest.raises(ChanGoError, match="not available"):
            scanner.get_version("1.4")

    @pytest.mark.parametrize(
        "version",
        [
            "1.1",
            Version("1.1", dtm.date(2024, 1, 1)),
            "1.2",
            Version("1.2", dtm.date(2024, 1, 2)),
            "1.3",
            Version("1.3", dtm.date(2024, 1, 3)),
            "1.3.1",
            Version("1.3.1", dtm.date(2024, 1, 3)),
            None,
        ],
    )
    def test_get_changes(self, scanner, version):
        changes = set(scanner.get_changes(version))

        uid = version.uid if isinstance(version, Version) else (version or "ur")

        assert changes == {f"uid_{uid.replace('.', '-')}_{idx}" for idx in range(3)}

    def test_get_changes_not_found(self, scanner):
        with pytest.raises(ChanGoError, match="not available"):
            scanner.get_changes("1.4")

    def test_invalidate_caches(self, scanner):
        original_versions = {
            Version("1.1", dtm.date(2024, 1, 1)),
            Version("1.2", dtm.date(2024, 1, 2)),
            Version("1.3", dtm.date(2024, 1, 3)),
            Version("1.3.1", dtm.date(2024, 1, 3)),
        }

        assert set(scanner.get_available_versions()) == original_versions

        new_directory = Path(self.DATA_ROOT / "1.4_2024-01-04")
        try:
            new_directory.mkdir()
            assert set(scanner.get_available_versions()) == original_versions
            scanner.invalidate_caches()
            assert set(scanner.get_available_versions()) == original_versions | {
                Version("1.4", dtm.date(2024, 1, 4))
            }
        finally:
            new_directory.rmdir()



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
"""This module contains error classes specific to the :mod:`chango` package."""

__all__ = ["ChanGoError", "UnsupportedMarkupError", "ValidationError"]


class ChanGoError(Exception):
    """Base class for all exceptions defined by the chango package."""


class ValidationError(ChanGoError):
    """Exception raised when a validation error occurs."""


class UnsupportedMarkupError(ChanGoError):
    """Exception raised when an unsupported markup is encountered."""



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
"""This module contains functionality that allows automatically rendering changelogs in
`Sphinx <https://https://www.sphinx-doc.org/>`_ documentation using ``chango``.

.. seealso:: :ref:`sphinx_ext`
"""

import typing
from pathlib import Path
from types import NoneType

from sphinx.application import Sphinx

from chango import __version__

__all__ = ["setup"]

from ._util import directive_factory


def setup(app: Sphinx) -> dict[str, typing.Any]:
    """Sets up the ``chango`` Sphinx extension.
    This currently does two things:

    1. Adds the ``chango`` directive to Sphinx, which allows you to include changelogs in your
       documentation.
    2. Adds a configuration value ``chango_pyproject_toml_path`` to the Sphinx configuration, which
       allows you to specify the path to the ``pyproject.toml`` file that contains the chango
       configuration.

    Args:
        app (:class:`sphinx.application.Sphinx`): The Sphinx application object.

    Returns:
        dict[:class:`str`, :class:`typing.Any`]: A dictionary containing metadata about the
            extension.
    """
    app.add_config_value(
        "chango_pyproject_toml_path",
        None,
        rebuild="env",
        # Path & PurePath do not work well with how sphinx handles config value type checks
        types=(str, type(Path.cwd()), NoneType),
        description=(
            "Path to the pyproject.toml file to use for the chango configuration. Takes "
            "the same inputs as `chango.config.ChanGoConfig.load`."
        ),
    )
    app.add_directive("chango", directive_factory(app))

    return {"version": __version__, "parallel_read_safe": True, "parallel_write_safe": True}



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
"""This module contains constants used throughout the :mod:`chango` package."""

__all__ = ["MarkupLanguage"]

import contextlib
from collections.abc import Mapping
from enum import StrEnum


class MarkupLanguage(StrEnum):
    """Commonly known markup languages"""

    ASCIIDOC = "asciidoc"
    """The `AsciiDoc <https://asciidoc.org>`_ markup language"""
    CREOLE = "creole"
    """The `Creole <https://www.wikicreole.org>`_ markup language"""
    HTML = "html"
    """The `HyperText Markup Language <https://html.spec.whatwg.org>`_"""
    MARKDOWN = "markdown"
    """The `Markdown <https://daringfireball.net/projects/markdown/>`_ markup language"""
    MEDIAWIKI = "mediawiki"
    """The `MediaWiki <https://www.mediawiki.org>`_ markup language"""
    ORG = "org"
    """The `Org-mode <https://orgmode.org>`_ markup language"""
    POD = "pod"
    """The `Plain Old Documentation <https://perldoc.perl.org/perlpod.html>`_ markup language"""
    RDOC = "rdoc"
    """The `RDoc <https://ruby.github.io/rdoc/>`_ markup language"""
    RESTRUCTUREDTEXT = "rst"
    """The `reStructuredText <https://docutils.sourceforge.io/rst.html>`_ markup language"""
    TEXTILE = "textile"
    """The `Textile <https://textile-lang.com>`_ markup language"""
    TEXT = "txt"
    """Plain text"""

    @classmethod
    def from_string(
        cls, string: str, mapping: Mapping[str, "MarkupLanguage"] | None = None
    ) -> "MarkupLanguage":
        """Get the markup language enum member from a string by comparing against the members of
        this enum as well as commonly used file extensions. Case-insensitive. Leading dots are
        ignored.

        Args:
            string (:obj:`str`): The string to look up.
            mapping (:class:`~collections.abc.Mapping` [:obj:`str`, :class:`MarkupLanguage`] | \
                :obj:`None`): A mapping of file extensions to markup languages. If not provided,
                the default mapping will be used.

        Returns:
            :class:`MarkupLanguage`: The markup language enum member.

        Raises:
            ValueError: If the file extension can not be resolved to a markup language.
        """
        lookup = string.lower().lstrip(".")

        with contextlib.suppress(ValueError):
            return cls(lookup)

        with contextlib.suppress(KeyError):
            return cls[lookup.upper()]

        effective_mapping = mapping or {
            "adoc": cls.ASCIIDOC,
            "htm": cls.HTML,
            "md": cls.MARKDOWN,
            "mkd": cls.MARKDOWN,
            "mdwn": cls.MARKDOWN,
            "mdown": cls.MARKDOWN,
            "mdtxt": cls.MARKDOWN,
            "mdtext": cls.MARKDOWN,
            "mediawiki": cls.MEDIAWIKI,
            "org": cls.ORG,
            "pod": cls.POD,
            "rdoc": cls.RDOC,
            "text": cls.TEXT,
        }

        with contextlib.suppress(KeyError):
            return effective_mapping[lookup]

        raise ValueError(f"File extension `{string}` not found in mapping.")



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import os
from contextlib import contextmanager
from pathlib import Path

from chango._utils.types import PathLike

PROJECT_ROOT_PATH = Path(__file__).parent.parent.parent.resolve()
TEST_DATA_PATH = PROJECT_ROOT_PATH / "tests" / "data"


def data_path(filename: PathLike) -> Path:
    return TEST_DATA_PATH / filename


def path_to_python_string(path: Path, output_type: type[str] | type[Path]) -> str:
    if output_type is str:
        return f"'{path.as_posix()}'"
    return f"Path(r'{path}')"


@contextmanager
def temporary_chdir(path: Path):
    current_dir = Path.cwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(current_dir)



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
from pathlib import Path

import pytest
import shortuuid

from chango.abc import ChangeNote
from chango.concrete import CommentChangeNote
from chango.error import ValidationError
from tests.auxil.files import data_path

UTF_8_PATH = data_path("comment-change-note.uid.txt")
UTF_16_PATH = data_path("comment-change-note-utf16.uid.txt")


@pytest.fixture(params=["string", "path"])
def utf_8_path(request):
    if request.param == "string":
        return str(UTF_8_PATH)
    return UTF_8_PATH


@pytest.fixture(params=["string", "path"])
def utf_16_path(request):
    if request.param == "string":
        return str(UTF_16_PATH)
    return UTF_16_PATH


class TestChangeNote:
    """Since ChangeNote is an abstract base class, we are testing with CommentChangeNote as a
    simple implementation.

    Note that we do *not* test abstract methods, as that is the responsibility of the concrete
    implementations.
    """

    change_note = CommentChangeNote(slug="slug", comment="this is a comment", uid="uid")

    def test_init(self):
        assert self.change_note.slug == "slug"
        assert self.change_note.uid == "uid"

        change_note = CommentChangeNote(slug="slug", comment="this is a comment")
        assert change_note.slug == "slug"
        assert isinstance(change_note.uid, str)
        assert len(change_note.uid) == len(shortuuid.ShortUUID().uuid())

    def test_init_invalid_slug(self):
        with pytest.raises(ValidationError, match="slug must not contain"):
            CommentChangeNote(slug="slug.with.dot", comment="this is a comment")

    def test_file_name(self):
        assert self.change_note.file_name == "slug.uid.txt"

    def test_from_file(self, utf_8_path):
        change_note = CommentChangeNote.from_file(utf_8_path)
        assert change_note.slug == "comment-change-note"
        assert change_note.uid == "uid"

    def test_from_file_encoding(self, utf_16_path):
        change_note = CommentChangeNote.from_file(utf_16_path, encoding="utf-16")
        assert change_note.slug == "comment-change-note-utf16"
        assert change_note.uid == "uid"
        assert change_note.comment == "this is an utf-16 comment 𝛙𝌢𑁍"

        with pytest.raises(UnicodeDecodeError):
            CommentChangeNote.from_file(utf_16_path, encoding="utf-8")

    def test_from_bytes(self):
        change_note = CommentChangeNote.from_bytes(
            slug="slug", uid="uid", data=UTF_8_PATH.read_bytes()
        )
        assert change_note.slug == "slug"
        assert change_note.uid == "uid"
        assert change_note.comment == "this is a comment"

    def test_from_bytes_encoding(self):
        change_note = CommentChangeNote.from_bytes(
            slug="slug", uid="uid", data=UTF_16_PATH.read_bytes(), encoding="utf-16"
        )
        assert change_note.slug == "slug"
        assert change_note.uid == "uid"
        assert change_note.comment == "this is an utf-16 comment 𝛙𝌢𑁍"

        with pytest.raises(UnicodeDecodeError):
            CommentChangeNote.from_bytes(
                slug="slug", uid="uid", data=UTF_16_PATH.read_bytes(), encoding="utf-8"
            )

    def test_to_bytes(self):
        assert self.change_note.to_bytes() == b"this is a comment"

    def test_to_bytes_encoding(self):
        change_note = CommentChangeNote(slug="slug", comment="this is a comment 𝛙𝌢𑁍", uid="uid")
        assert change_note.to_bytes(encoding="utf-16") == "this is a comment 𝛙𝌢𑁍".encode("utf-16")

    @pytest.mark.parametrize("directory", [None, "custom"])
    def test_to_file(self, tmp_path, directory):
        path = None
        expected_dir = tmp_path if directory == "custom" else Path.cwd()
        try:
            path = self.change_note.to_file(directory=tmp_path if directory == "custom" else None)
            assert path == expected_dir / "slug.uid.txt"
            assert path.read_text() == "this is a comment"
        finally:
            if path:
                path.unlink()

    @pytest.mark.parametrize("directory", [None, "custom"])
    def test_to_file_encoding(self, tmp_path, directory):
        path = None
        expected_dir = tmp_path if directory == "custom" else Path.cwd()
        change_note = CommentChangeNote(slug="slug", comment="this is a comment 𝛙𝌢𑁍", uid="uid")

        try:
            path = change_note.to_file(
                directory=tmp_path if directory == "custom" else None, encoding="utf-16"
            )
            assert path == expected_dir / "slug.uid.txt"
            assert path.read_text(encoding="utf-16") == "this is a comment 𝛙𝌢𑁍"

            with pytest.raises(UnicodeDecodeError):
                path.read_text(encoding="utf-8")
        finally:
            if path:
                path.unlink()

    def test_build_from_github_event(self, monkeypatch):
        monkeypatch.setattr(
            self.change_note, "build_from_github_event", ChangeNote.build_from_github_event
        )
        with pytest.raises(NotImplementedError):
            self.change_note.build_from_github_event({})

    def test_update_uid(self):
        change_note = CommentChangeNote(slug="slug", comment="this is a comment", uid="uid")
        assert change_note.uid == "uid"
        change_note.update_uid("abc")
        assert change_note.uid == "abc"
        assert change_note.slug == "slug"
        assert change_note.file_name == "slug.abc.txt"



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import datetime as dtm
from pathlib import Path
from typing import Annotated

import typer

from chango.constants import MarkupLanguage


def markup_callback(value: str) -> MarkupLanguage:
    try:
        return MarkupLanguage.from_string(value)
    except ValueError as exc:
        raise typer.BadParameter(str(exc)) from exc


def date(value: str | dtm.date) -> dtm.date:
    if isinstance(value, dtm.date):
        return value
    try:
        return dtm.date.fromisoformat(value)
    except ValueError as exc:
        raise typer.BadParameter(str(exc)) from exc


MARKUP = Annotated[
    str,
    typer.Option(
        ...,
        "-m",
        "--markup",
        help="The markup language to use for the report.",
        callback=markup_callback,
    ),
]
OUTPUT_FILE = Annotated[
    Path | None,
    typer.Option(
        ...,
        "-o",
        "--output",
        help="The file to write to. If not specified, the output is printed to the console.",
        dir_okay=False,
        writable=True,
    ),
]



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
from typing import ClassVar

import pydantic as pydt


class Section(pydt.BaseModel):
    """Configuration for a section in a :class:`SectionChangeNote`.

    Args:
        uid (:obj:`str`): The unique identifier for the section. This is used as the field name
            in the change note.
        title (:obj:`str`): The title of the section.
        is_required (:obj:`bool`, optional): Whether the section is required. Defaults
            to :obj:`False`.

            Tip:
                At least one section must be required.
        render_pr_details (:obj:`bool`, optional): Whether to include details about the pull
            requests related to the change in the rendering for this section.
            Defaults to :obj:`True`.
        sort_order (:obj:`int`, optional): The sort order of the section. Defaults to ``0``.

    Attributes:
        uid (:obj:`str`): The unique identifier for the section.
        title (:obj:`str`): The title of the section.
        is_required (:obj:`bool`): Whether the section is required.
        render_pr_details (:obj:`bool`, optional): Whether to include details about the pull
            requests related to the change in the rendering for this section.
        sort_order (:obj:`int`): The sort order of the section.

    """

    model_config: ClassVar[pydt.ConfigDict] = pydt.ConfigDict(frozen=True)

    uid: str
    title: str
    is_required: bool = False
    render_pr_details: bool = True
    sort_order: int = 0



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import pytest

from chango.action import ChanGoActionData, LinkedIssue
from chango.concrete.sections import GitHubSectionChangeNote, PullRequest, Section


class DummyChangNote(
    GitHubSectionChangeNote.with_sections([Section(uid="req", title="Req", is_required=True)])
):
    OWNER = "my-username"
    REPOSITORY = "my-repo"


class DummyChangNoteNoOwner(
    GitHubSectionChangeNote.with_sections([Section(uid="req", title="Req", is_required=True)])
):
    REPOSITORY = "my-repo"


class DummyChangNoteNoRepository(
    GitHubSectionChangeNote.with_sections([Section(uid="req", title="Req", is_required=True)])
):
    OWNER = "my-username"


class FromGitHubEvent(
    GitHubSectionChangeNote.with_sections(
        [
            Section(uid="opt_0", title="Opt", is_required=False, sort_order=0),
            Section(uid="req_1", title="Req", is_required=True, sort_order=1),
            Section(uid="req_0", title="Req", is_required=True, sort_order=0),
        ]
    )
):
    pass


class TestGitHubSectionChangeNote:
    """Since TestSectionChangeNote is an abstract base class, we are testing with
    GitHubTestSectionChangeNote as a simple implementation.

    Note that we do *not* test abstract methods, as that is the responsibility of the concrete
    implementations.
    """

    def test_class_variables(self):
        assert DummyChangNote.OWNER == "my-username"
        assert DummyChangNote.REPOSITORY == "my-repo"

    def test_get_pull_request_url(self):
        assert (
            DummyChangNote.get_pull_request_url("123")
            == "https://github.com/my-username/my-repo/pull/123"
        )

    def test_get_pull_request_url_invalid(self):
        with pytest.raises(ValueError, match=r"OWNER must be set as class variable."):
            DummyChangNoteNoOwner.get_pull_request_url("123")
        with pytest.raises(ValueError, match=r"REPOSITORY must be set as class variable."):
            DummyChangNoteNoRepository.get_pull_request_url("123")

    def test_get_thread_url(self):
        assert (
            DummyChangNote.get_thread_url("123")
            == "https://github.com/my-username/my-repo/issues/123"
        )

    def test_get_thread_url_invalid(self):
        with pytest.raises(ValueError, match=r"OWNER must be set as class variable."):
            DummyChangNoteNoOwner.get_thread_url("123")
        with pytest.raises(ValueError, match=r"REPOSITORY must be set as class variable."):
            DummyChangNoteNoRepository.get_thread_url("123")

    def test_get_author_url(self):
        assert DummyChangNote.get_author_url("123") == "https://github.com/123"

    def test_get_sections_has_required(self):
        assert FromGitHubEvent.get_sections(None, None) == {"req_0", "req_1"}

    def test_get_sections_no_required(self):
        NoRequired = GitHubSectionChangeNote.with_sections(
            [
                Section(uid="opt_0", title="Opt", is_required=False, sort_order=5),
                Section(uid="opt_1", title="Opt", is_required=False, sort_order=42),
                Section(uid="opt_2", title="Opt", is_required=False, sort_order=-3),
            ]
        )
        assert NoRequired.get_sections(None, None) == {"opt_2"}

    def test_build_from_github_event_missing_data(self):
        with pytest.raises(ValueError, match="required data"):
            FromGitHubEvent.build_from_github_event({})

    def test_build_from_github_event_basic(self):
        event_data = {
            "pull_request": {
                "html_url": "https://example.com/pull/42",
                "number": 42,
                "title": "example title",
                "user": {"login": "author_uid"},
                "labels": [{"name": "label1"}, {"name": "label2"}],
            }
        }

        change_note = FromGitHubEvent.build_from_github_event(event_data)
        assert change_note.req_0 == "example title"
        assert change_note.pull_requests == (
            PullRequest(uid="42", author_uids=("author_uid",), closes_threads=()),
        )
        assert change_note.slug == "0042"

    def test_build_from_github_event_custom_get_sections(self):
        class CustomGetSections(FromGitHubEvent):
            @classmethod
            def get_sections(cls, event_data, sections):  # noqa: ARG003
                return {"opt_0", "req_1", "req_0"}

        event_data = {
            "pull_request": {
                "html_url": "https://example.com/pull/42",
                "number": 42,
                "title": "example title",
                "user": {"login": "author_uid"},
                "labels": [{"name": "label1"}, {"name": "label2"}],
            }
        }

        change_note = CustomGetSections.build_from_github_event(event_data)
        assert change_note.req_0 == "example title"
        assert change_note.req_1 == "example title"
        assert change_note.opt_0 == "example title"
        assert change_note.pull_requests == (
            PullRequest(uid="42", author_uids=("author_uid",), closes_threads=()),
        )
        assert change_note.slug == "0042"

    def test_build_from_github_event_chango_action_data_no_linked_issues(self, monkeypatch):
        received_data = {}

        def get_sections(labels, issue_types):
            received_data["labels"] = labels
            received_data["issue_types"] = issue_types
            return {"req_0", "req_1"}

        monkeypatch.setattr(FromGitHubEvent, "get_sections", get_sections)

        event_data = {
            "pull_request": {
                "html_url": "https://example.com/pull/42",
                "number": 42,
                "title": "example title",
                "user": {"login": "author_uid"},
                "labels": [],
            }
        }
        data = ChanGoActionData(linked_issues=None, parent_pull_request=None)

        FromGitHubEvent.build_from_github_event(event_data, data)
        assert received_data["labels"] == set()
        assert received_data["issue_types"] == set()

    def test_build_from_github_event_chango_action_data(self, monkeypatch):
        received_data = {}

        def get_sections(labels, issue_types):
            received_data["labels"] = labels
            received_data["issue_types"] = issue_types
            return {"req_0", "req_1"}

        monkeypatch.setattr(FromGitHubEvent, "get_sections", get_sections)

        event_data = {
            "pull_request": {
                "html_url": "https://example.com/pull/42",
                "number": 42,
                "title": "example title",
                "user": {"login": "author_uid"},
                "labels": [{"name": "pr_label1"}, {"name": "pr_label2"}],
            }
        }
        data = ChanGoActionData(
            linked_issues=(
                LinkedIssue(number=1, title="issue_title", labels=None, issue_type="issue_type"),
                LinkedIssue(
                    number=2,
                    title="issue_title",
                    labels=("issue_label1", "issue_label2"),
                    issue_type=None,
                ),
            ),
            parent_pull_request=None,
        )

        FromGitHubEvent.build_from_github_event(event_data, data)
        assert received_data["labels"] == {
            "pr_label1",
            "pr_label2",
            "issue_label1",
            "issue_label2",
        }
        assert received_data["issue_types"] == {"issue_type"}



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
import datetime as dtm
from pathlib import Path

import pytest

from chango import Version
from chango.concrete import CommentChangeNote, CommentVersionNote, HeaderVersionHistory
from chango.constants import MarkupLanguage
from chango.error import UnsupportedMarkupError
from tests.auxil.files import data_path


class TestHeaderVersionHistory:
    comments = ("comment 1", "a\nmulti-line\ncomment 2", "comment 3")

    @staticmethod
    def get_expected_file(unreleased_changes: bool, markup: MarkupLanguage) -> Path:
        unreleased_prefix = "with-unreleased" if unreleased_changes else "without-unreleased"
        file_name = f"{unreleased_prefix}.{markup}"
        return data_path(Path("headerversionhistory") / file_name)

    @classmethod
    def get_expected_string(cls, unreleased_changes: bool, markup: MarkupLanguage) -> str:
        return cls.get_expected_file(unreleased_changes, markup).read_text()

    def get_version_notes(self, unreleased_changes: bool) -> list[CommentVersionNote]:
        version_notes = []
        for j, version_number in enumerate(range(3)):
            version = Version(uid=f"1.0.{version_number}", date=dtm.date(2024, 1, 1 + j))
            version_note = CommentVersionNote(version=version)
            version_notes.append(version_note)

            for i, comment in enumerate(self.comments):
                change_note = CommentChangeNote(slug=f"slug-{i}", comment=comment)
                version_note.add_change_note(change_note)

        if not unreleased_changes:
            return version_notes

        unreleased_version_note = CommentVersionNote(version=None)
        version_notes.append(unreleased_version_note)
        for i, comment in enumerate(self.comments):
            change_note = CommentChangeNote(slug=f"slug-{i}", comment=comment)
            unreleased_version_note.add_change_note(change_note)

        return version_notes

    @pytest.mark.parametrize(
        "markup", [MarkupLanguage.MARKDOWN, MarkupLanguage.HTML, MarkupLanguage.RESTRUCTUREDTEXT]
    )
    @pytest.mark.parametrize(
        "unreleased_changes", [False, True], ids=["without-unreleased", "with-unreleased"]
    )
    def test_expected_output(self, unreleased_changes: bool, markup: MarkupLanguage):
        version_notes = self.get_version_notes(unreleased_changes=unreleased_changes)
        version_history = HeaderVersionHistory()
        for version_note in version_notes:
            version_history.add_version_note(version_note)
        expected = self.get_expected_string(unreleased_changes=unreleased_changes, markup=markup)
        assert version_history.render(markup) == expected

    def test_unsupported_markup(self):
        version_history = HeaderVersionHistory()
        with pytest.raises(UnsupportedMarkupError, match="Got unsupported markup 'unsupported'"):
            version_history.render("unsupported")



#  SPDX-FileCopyrightText: 2024-present Hinrich Mahler <chango@mahlerhome.de>
#
#  SPDX-License-Identifier: MIT
"""This module contains functionality required when using chango in :ref:`action`."""

from typing import ClassVar, Literal

from pydantic import AnyHttpUrl, BaseModel, ConfigDict

__all__ = ["ChanGoActionData", "LinkedIssue", "ParentPullRequest"]


class _FrozenModel(BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(frozen=True)


class ParentPullRequest(_FrozenModel):
    """Data structure for a pull request associated with the target branch of the current pull
    request.

    Args:
        number (:obj:`int`): The pull request number.
        title (:obj:`str`): The title of the pull request.
        url (:obj:`str`): The URL of the pull request.
        state (:obj:`str`): The state of the pull request. Possible values are ``open``,
            ``closed``, and ``merged``.

    Attributes:
        number (:obj:`int`): The pull request number.
        author_login (:obj:`str`): The login of the author of the pull request.
        title (:obj:`str`): The title of the pull request.
        url (:obj:`str`): The URL of the pull request.
        state (:obj:`str`): The state of the pull request. Possible values are ``OPEN``,
            ``CLOSED``, and ``MERGED``.
    """

    number: int
    author_login: str
    title: str
    url: AnyHttpUrl
    state: Literal["OPEN", "CLOSED", "MERGED"]


class LinkedIssue(_FrozenModel):
    """Data structure for an issue linked in a GitHub pull request.

    Args:
        number (:obj:`int`): The issue number.
        title (:obj:`str`): The title of the issue.
        labels (tuple[:obj:`str`], optional): The labels of the issue.
        issue_type (:obj:`str`, optional): The type of the issue.

    Attributes:
        number (:obj:`int`): The issue number.
        title (:obj:`str`): The title of the issue.
        labels (tuple[:obj:`str`]): Optional. The labels of the issue.
        issue_type (:obj:`str`): Optional. The type of the issue.
    """

    number: int
    title: str
    labels: tuple[str, ...] | None
    issue_type: str | None = None


class ChanGoActionData(_FrozenModel):
    """Data structure for the additional information that the ``chango`` action automatically
    provides in addition to the GitHub event payload.

    Args:
        parent_pull_request (:class:`ParentPullRequest` | :obj:`None`): If there is a pull request
            associated with the target branch of the current pull request, this field contains
            information about it.
        linked_issues (tuple[:class:`LinkedIssue`], optional): Information about linked issues,
            i.e., issues that will be closed when the current pull request is merged.

    Attributes:
        parent_pull_request (:class:`ParentPullRequest`): Optional. If there is a pull request
            associated with the target branch of the current pull request, this field contains
            information about it.
        linked_issues (tuple[:class:`LinkedIssue`]): Optional. Information about linked issues,
            i.e., issues that will be closed when the current pull request is merged.
    """

    parent_pull_request: ParentPullRequest | None
    linked_issues: tuple[LinkedIssue, ...] | None
